"""
Enterprise Event Store Module.

Provides event store abstraction, event persistence,
snapshots, and event sourcing support.

Example:
    # Create event store
    store = create_event_store()
    
    # Append events
    await store.append(
        stream_id="order-123",
        events=[OrderCreated(order_id="123")],
    )
    
    # Read events
    events = await store.read("order-123")
"""

from __future__ import annotations

import asyncio
import json
import logging
import uuid
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import (
    Any,
    AsyncIterator,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Type,
    TypeVar,
)

T = TypeVar('T')
E = TypeVar('E', bound='DomainEvent')


logger = logging.getLogger(__name__)


class EventStoreError(Exception):
    """Event store error."""
    pass


class ConcurrencyError(EventStoreError):
    """Optimistic concurrency error."""
    pass


class StreamNotFoundError(EventStoreError):
    """Stream not found error."""
    pass


class EventReadDirection(str, Enum):
    """Event read direction."""
    FORWARD = "forward"
    BACKWARD = "backward"


class StreamState(str, Enum):
    """Stream state."""
    ACTIVE = "active"
    DELETED = "deleted"
    ARCHIVED = "archived"


@dataclass
class EventMetadata:
    """Event metadata."""
    event_id: str = field(default_factory=lambda: str(uuid.uuid4()))
    correlation_id: Optional[str] = None
    causation_id: Optional[str] = None
    timestamp: datetime = field(default_factory=datetime.now)
    user_id: Optional[str] = None
    custom: Dict[str, Any] = field(default_factory=dict)


@dataclass
class DomainEvent:
    """Base domain event."""
    event_type: str = ""
    
    def __post_init__(self):
        if not self.event_type:
            self.event_type = type(self).__name__


@dataclass
class RecordedEvent:
    """Recorded event with metadata."""
    stream_id: str
    event_number: int
    event_type: str
    data: Any
    metadata: EventMetadata
    global_position: int = 0
    
    @property
    def event_id(self) -> str:
        return self.metadata.event_id
    
    @property
    def timestamp(self) -> datetime:
        return self.metadata.timestamp


@dataclass
class StreamMetadata:
    """Stream metadata."""
    stream_id: str
    version: int = 0
    state: StreamState = StreamState.ACTIVE
    created_at: datetime = field(default_factory=datetime.now)
    updated_at: datetime = field(default_factory=datetime.now)
    custom: Dict[str, Any] = field(default_factory=dict)


@dataclass
class EventSlice:
    """Slice of events from a stream."""
    events: List[RecordedEvent]
    from_version: int
    to_version: int
    is_end: bool = False
    next_version: Optional[int] = None


@dataclass
class AppendResult:
    """Result of appending events."""
    stream_id: str
    version: int
    events_count: int
    global_positions: List[int]


@dataclass
class Snapshot:
    """Aggregate snapshot."""
    stream_id: str
    version: int
    state: Any
    timestamp: datetime = field(default_factory=datetime.now)
    metadata: Dict[str, Any] = field(default_factory=dict)


class EventSerializer(ABC):
    """
    Event serializer interface.
    """
    
    @abstractmethod
    def serialize(self, event: Any) -> bytes:
        """Serialize event to bytes."""
        pass
    
    @abstractmethod
    def deserialize(self, data: bytes, event_type: str) -> Any:
        """Deserialize bytes to event."""
        pass


class JsonEventSerializer(EventSerializer):
    """
    JSON event serializer.
    """
    
    def __init__(self):
        self._type_map: Dict[str, Type] = {}
    
    def register_type(self, event_type: str, cls: Type) -> None:
        """Register event type mapping."""
        self._type_map[event_type] = cls
    
    def serialize(self, event: Any) -> bytes:
        if hasattr(event, '__dict__'):
            data = event.__dict__.copy()
        else:
            data = event
        return json.dumps(data, default=str).encode('utf-8')
    
    def deserialize(self, data: bytes, event_type: str) -> Any:
        parsed = json.loads(data.decode('utf-8'))
        
        if event_type in self._type_map:
            cls = self._type_map[event_type]
            return cls(**parsed)
        
        return parsed


class EventStore(ABC):
    """
    Abstract event store.
    """
    
    @abstractmethod
    async def append(
        self,
        stream_id: str,
        events: List[Any],
        expected_version: Optional[int] = None,
        metadata: Optional[EventMetadata] = None,
    ) -> AppendResult:
        """Append events to a stream."""
        pass
    
    @abstractmethod
    async def read(
        self,
        stream_id: str,
        from_version: int = 0,
        max_count: int = 1000,
        direction: EventReadDirection = EventReadDirection.FORWARD,
    ) -> EventSlice:
        """Read events from a stream."""
        pass
    
    @abstractmethod
    async def read_all(
        self,
        from_position: int = 0,
        max_count: int = 1000,
        direction: EventReadDirection = EventReadDirection.FORWARD,
    ) -> EventSlice:
        """Read all events across streams."""
        pass
    
    @abstractmethod
    async def get_stream_metadata(
        self,
        stream_id: str,
    ) -> Optional[StreamMetadata]:
        """Get stream metadata."""
        pass
    
    @abstractmethod
    async def delete_stream(
        self,
        stream_id: str,
        expected_version: Optional[int] = None,
    ) -> None:
        """Delete a stream."""
        pass


class InMemoryEventStore(EventStore):
    """
    In-memory event store implementation.
    """
    
    def __init__(self, serializer: Optional[EventSerializer] = None):
        self._streams: Dict[str, List[RecordedEvent]] = defaultdict(list)
        self._metadata: Dict[str, StreamMetadata] = {}
        self._global_position = 0
        self._serializer = serializer or JsonEventSerializer()
        self._subscribers: List[Callable] = []
        self._lock = asyncio.Lock()
    
    async def append(
        self,
        stream_id: str,
        events: List[Any],
        expected_version: Optional[int] = None,
        metadata: Optional[EventMetadata] = None,
    ) -> AppendResult:
        async with self._lock:
            stream = self._streams[stream_id]
            current_version = len(stream)
            
            # Optimistic concurrency check
            if expected_version is not None:
                if expected_version != current_version:
                    raise ConcurrencyError(
                        f"Expected version {expected_version}, "
                        f"but stream is at version {current_version}"
                    )
            
            global_positions = []
            base_metadata = metadata or EventMetadata()
            
            for i, event in enumerate(events):
                self._global_position += 1
                
                event_metadata = EventMetadata(
                    event_id=str(uuid.uuid4()),
                    correlation_id=base_metadata.correlation_id,
                    causation_id=base_metadata.causation_id,
                    user_id=base_metadata.user_id,
                )
                
                event_type = (
                    event.event_type
                    if hasattr(event, 'event_type')
                    else type(event).__name__
                )
                
                recorded = RecordedEvent(
                    stream_id=stream_id,
                    event_number=current_version + i,
                    event_type=event_type,
                    data=event,
                    metadata=event_metadata,
                    global_position=self._global_position,
                )
                
                stream.append(recorded)
                global_positions.append(self._global_position)
            
            # Update stream metadata
            if stream_id not in self._metadata:
                self._metadata[stream_id] = StreamMetadata(
                    stream_id=stream_id,
                )
            
            self._metadata[stream_id].version = len(stream)
            self._metadata[stream_id].updated_at = datetime.now()
            
            # Notify subscribers
            for subscriber in self._subscribers:
                for recorded in stream[-len(events):]:
                    try:
                        if asyncio.iscoroutinefunction(subscriber):
                            await subscriber(recorded)
                        else:
                            subscriber(recorded)
                    except Exception as e:
                        logger.error(f"Subscriber error: {e}")
            
            return AppendResult(
                stream_id=stream_id,
                version=len(stream),
                events_count=len(events),
                global_positions=global_positions,
            )
    
    async def read(
        self,
        stream_id: str,
        from_version: int = 0,
        max_count: int = 1000,
        direction: EventReadDirection = EventReadDirection.FORWARD,
    ) -> EventSlice:
        stream = self._streams.get(stream_id, [])
        
        if direction == EventReadDirection.FORWARD:
            events = stream[from_version:from_version + max_count]
            to_version = from_version + len(events) - 1 if events else from_version
            is_end = from_version + max_count >= len(stream)
            next_version = (
                from_version + len(events)
                if not is_end
                else None
            )
        else:
            start = max(0, from_version - max_count + 1)
            events = list(reversed(stream[start:from_version + 1]))
            to_version = start
            is_end = start == 0
            next_version = start - 1 if not is_end else None
        
        return EventSlice(
            events=events,
            from_version=from_version,
            to_version=to_version,
            is_end=is_end,
            next_version=next_version,
        )
    
    async def read_all(
        self,
        from_position: int = 0,
        max_count: int = 1000,
        direction: EventReadDirection = EventReadDirection.FORWARD,
    ) -> EventSlice:
        # Collect all events across streams
        all_events: List[RecordedEvent] = []
        for events in self._streams.values():
            all_events.extend(events)
        
        # Sort by global position
        all_events.sort(key=lambda e: e.global_position)
        
        if direction == EventReadDirection.FORWARD:
            filtered = [
                e for e in all_events
                if e.global_position >= from_position
            ][:max_count]
        else:
            filtered = [
                e for e in all_events
                if e.global_position <= from_position
            ][-max_count:]
            filtered.reverse()
        
        return EventSlice(
            events=filtered,
            from_version=from_position,
            to_version=(
                filtered[-1].global_position
                if filtered
                else from_position
            ),
            is_end=len(filtered) < max_count,
        )
    
    async def get_stream_metadata(
        self,
        stream_id: str,
    ) -> Optional[StreamMetadata]:
        return self._metadata.get(stream_id)
    
    async def delete_stream(
        self,
        stream_id: str,
        expected_version: Optional[int] = None,
    ) -> None:
        if stream_id not in self._streams:
            return
        
        if expected_version is not None:
            current_version = len(self._streams[stream_id])
            if expected_version != current_version:
                raise ConcurrencyError(
                    f"Expected version {expected_version}, "
                    f"but stream is at version {current_version}"
                )
        
        if stream_id in self._metadata:
            self._metadata[stream_id].state = StreamState.DELETED
        
        del self._streams[stream_id]
    
    def subscribe(
        self,
        handler: Callable[[RecordedEvent], Any],
    ) -> Callable[[], None]:
        """Subscribe to new events."""
        self._subscribers.append(handler)
        
        def unsubscribe():
            self._subscribers.remove(handler)
        
        return unsubscribe


class SnapshotStore(ABC):
    """
    Abstract snapshot store.
    """
    
    @abstractmethod
    async def save(self, snapshot: Snapshot) -> None:
        """Save a snapshot."""
        pass
    
    @abstractmethod
    async def load(
        self,
        stream_id: str,
        max_version: Optional[int] = None,
    ) -> Optional[Snapshot]:
        """Load a snapshot."""
        pass
    
    @abstractmethod
    async def delete(self, stream_id: str) -> None:
        """Delete snapshots for a stream."""
        pass


class InMemorySnapshotStore(SnapshotStore):
    """
    In-memory snapshot store.
    """
    
    def __init__(self):
        self._snapshots: Dict[str, List[Snapshot]] = defaultdict(list)
    
    async def save(self, snapshot: Snapshot) -> None:
        self._snapshots[snapshot.stream_id].append(snapshot)
        # Keep only latest N snapshots
        self._snapshots[snapshot.stream_id] = (
            self._snapshots[snapshot.stream_id][-10:]
        )
    
    async def load(
        self,
        stream_id: str,
        max_version: Optional[int] = None,
    ) -> Optional[Snapshot]:
        snapshots = self._snapshots.get(stream_id, [])
        
        if not snapshots:
            return None
        
        if max_version is not None:
            valid = [s for s in snapshots if s.version <= max_version]
            return valid[-1] if valid else None
        
        return snapshots[-1]
    
    async def delete(self, stream_id: str) -> None:
        self._snapshots.pop(stream_id, None)


class EventStreamReader:
    """
    Helper for reading event streams.
    """
    
    def __init__(
        self,
        store: EventStore,
        stream_id: str,
    ):
        self._store = store
        self._stream_id = stream_id
    
    async def read_all(self) -> List[RecordedEvent]:
        """Read all events from stream."""
        events = []
        version = 0
        
        while True:
            slice = await self._store.read(
                self._stream_id,
                from_version=version,
            )
            events.extend(slice.events)
            
            if slice.is_end:
                break
            
            version = slice.next_version or 0
        
        return events
    
    async def __aiter__(self) -> AsyncIterator[RecordedEvent]:
        """Iterate over events."""
        version = 0
        
        while True:
            slice = await self._store.read(
                self._stream_id,
                from_version=version,
            )
            
            for event in slice.events:
                yield event
            
            if slice.is_end:
                break
            
            version = slice.next_version or 0


class EventSourcedRepository(Generic[T]):
    """
    Repository for event-sourced aggregates.
    """
    
    def __init__(
        self,
        event_store: EventStore,
        snapshot_store: Optional[SnapshotStore] = None,
        snapshot_interval: int = 100,
    ):
        self._event_store = event_store
        self._snapshot_store = snapshot_store
        self._snapshot_interval = snapshot_interval
        self._factory: Optional[Callable[[], T]] = None
        self._apply_map: Dict[str, Callable[[T, Any], None]] = {}
    
    def set_factory(self, factory: Callable[[], T]) -> None:
        """Set aggregate factory."""
        self._factory = factory
    
    def register_apply(
        self,
        event_type: str,
        handler: Callable[[T, Any], None],
    ) -> None:
        """Register event apply handler."""
        self._apply_map[event_type] = handler
    
    async def load(self, stream_id: str) -> Optional[T]:
        """Load aggregate from events."""
        if not self._factory:
            raise EventStoreError("No factory registered")
        
        # Try to load from snapshot
        snapshot = None
        from_version = 0
        
        if self._snapshot_store:
            snapshot = await self._snapshot_store.load(stream_id)
            if snapshot:
                aggregate = snapshot.state
                from_version = snapshot.version
            else:
                aggregate = self._factory()
        else:
            aggregate = self._factory()
        
        # Apply events
        slice = await self._event_store.read(
            stream_id,
            from_version=from_version,
        )
        
        if not slice.events and not snapshot:
            return None
        
        for event in slice.events:
            handler = self._apply_map.get(event.event_type)
            if handler:
                handler(aggregate, event.data)
        
        return aggregate
    
    async def save(
        self,
        stream_id: str,
        aggregate: T,
        events: List[Any],
        expected_version: Optional[int] = None,
    ) -> AppendResult:
        """Save aggregate with new events."""
        result = await self._event_store.append(
            stream_id,
            events,
            expected_version=expected_version,
        )
        
        # Create snapshot if needed
        if (
            self._snapshot_store
            and result.version % self._snapshot_interval == 0
        ):
            snapshot = Snapshot(
                stream_id=stream_id,
                version=result.version,
                state=aggregate,
            )
            await self._snapshot_store.save(snapshot)
        
        return result


class Subscription:
    """
    Event subscription for catching up with events.
    """
    
    def __init__(
        self,
        store: EventStore,
        handler: Callable[[RecordedEvent], Any],
        from_position: int = 0,
    ):
        self._store = store
        self._handler = handler
        self._position = from_position
        self._running = False
        self._task: Optional[asyncio.Task] = None
    
    async def start(self) -> None:
        """Start subscription."""
        if self._running:
            return
        
        self._running = True
        self._task = asyncio.create_task(self._run())
    
    async def stop(self) -> None:
        """Stop subscription."""
        self._running = False
        if self._task:
            self._task.cancel()
            try:
                await self._task
            except asyncio.CancelledError:
                pass
    
    async def _run(self) -> None:
        """Run subscription loop."""
        while self._running:
            try:
                slice = await self._store.read_all(
                    from_position=self._position,
                    max_count=100,
                )
                
                for event in slice.events:
                    try:
                        if asyncio.iscoroutinefunction(self._handler):
                            await self._handler(event)
                        else:
                            self._handler(event)
                        self._position = event.global_position + 1
                    except Exception as e:
                        logger.error(f"Subscription handler error: {e}")
                
                if slice.is_end:
                    await asyncio.sleep(0.1)
                    
            except Exception as e:
                logger.error(f"Subscription error: {e}")
                await asyncio.sleep(1.0)
    
    @property
    def position(self) -> int:
        """Get current position."""
        return self._position


class EventStoreRegistry:
    """
    Registry for event stores.
    """
    
    def __init__(self):
        self._stores: Dict[str, EventStore] = {}
        self._default: Optional[str] = None
    
    def register(
        self,
        name: str,
        store: EventStore,
        default: bool = False,
    ) -> None:
        """Register an event store."""
        self._stores[name] = store
        if default or self._default is None:
            self._default = name
    
    def get(self, name: Optional[str] = None) -> EventStore:
        """Get an event store."""
        name = name or self._default
        if not name or name not in self._stores:
            raise EventStoreError(f"Event store not found: {name}")
        return self._stores[name]


# Global registry
_global_registry = EventStoreRegistry()


# Decorators
def event_type(type_name: str) -> Callable:
    """
    Decorator to specify event type name.
    
    Example:
        @event_type("OrderCreated")
        class OrderCreated:
            order_id: str
    """
    def decorator(cls: Type) -> Type:
        cls.event_type = type_name
        return cls
    
    return decorator


def applies(event_type: str) -> Callable:
    """
    Decorator for event apply handlers.
    
    Example:
        @applies("OrderCreated")
        def apply_order_created(self, event):
            self.status = "created"
    """
    def decorator(func: Callable) -> Callable:
        func._applies_event_type = event_type
        return func
    
    return decorator


# Factory functions
def create_event_store() -> InMemoryEventStore:
    """Create an in-memory event store."""
    return InMemoryEventStore()


def create_snapshot_store() -> InMemorySnapshotStore:
    """Create an in-memory snapshot store."""
    return InMemorySnapshotStore()


def create_event_serializer() -> JsonEventSerializer:
    """Create a JSON event serializer."""
    return JsonEventSerializer()


def create_event_sourced_repository(
    event_store: Optional[EventStore] = None,
    snapshot_store: Optional[SnapshotStore] = None,
) -> EventSourcedRepository:
    """Create an event-sourced repository."""
    return EventSourcedRepository(
        event_store=event_store or create_event_store(),
        snapshot_store=snapshot_store,
    )


def create_subscription(
    store: EventStore,
    handler: Callable[[RecordedEvent], Any],
    from_position: int = 0,
) -> Subscription:
    """Create an event subscription."""
    return Subscription(store, handler, from_position)


def create_stream_reader(
    store: EventStore,
    stream_id: str,
) -> EventStreamReader:
    """Create a stream reader."""
    return EventStreamReader(store, stream_id)


def register_store(
    name: str,
    store: EventStore,
    default: bool = False,
) -> None:
    """Register store in global registry."""
    _global_registry.register(name, store, default)


def get_store(name: Optional[str] = None) -> EventStore:
    """Get store from global registry."""
    return _global_registry.get(name)


__all__ = [
    # Exceptions
    "EventStoreError",
    "ConcurrencyError",
    "StreamNotFoundError",
    # Enums
    "EventReadDirection",
    "StreamState",
    # Data classes
    "EventMetadata",
    "DomainEvent",
    "RecordedEvent",
    "StreamMetadata",
    "EventSlice",
    "AppendResult",
    "Snapshot",
    # Serializers
    "EventSerializer",
    "JsonEventSerializer",
    # Event Store
    "EventStore",
    "InMemoryEventStore",
    # Snapshot Store
    "SnapshotStore",
    "InMemorySnapshotStore",
    # Utilities
    "EventStreamReader",
    "EventSourcedRepository",
    "Subscription",
    # Registry
    "EventStoreRegistry",
    # Decorators
    "event_type",
    "applies",
    # Factory functions
    "create_event_store",
    "create_snapshot_store",
    "create_event_serializer",
    "create_event_sourced_repository",
    "create_subscription",
    "create_stream_reader",
    "register_store",
    "get_store",
]



"""
Enterprise Batch Processing Module.

Provides batch processing, job queues, and parallel execution
for handling large volumes of requests efficiently.

Example:
    # Batch processor
    processor = BatchProcessor(
        handler=process_item,
        batch_size=100,
        max_concurrent=5,
    )
    
    results = await processor.process(items)
    
    # Job queue
    queue = JobQueue(workers=4)
    job_id = await queue.submit(my_task, arg1, arg2)
    result = await queue.get_result(job_id)
"""

from __future__ import annotations

import asyncio
import uuid
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Sequence,
    TypeVar,
    Union,
)
from enum import Enum
import logging

logger = logging.getLogger(__name__)

T = TypeVar("T")
R = TypeVar("R")


class JobStatus(str, Enum):
    """Job execution status."""
    PENDING = "pending"
    RUNNING = "running"
    COMPLETED = "completed"
    FAILED = "failed"
    CANCELLED = "cancelled"
    RETRYING = "retrying"


class BatchError(Exception):
    """Batch processing error."""
    pass


class JobError(Exception):
    """Job execution error."""
    
    def __init__(self, message: str, job_id: str, original: Optional[Exception] = None):
        super().__init__(message)
        self.job_id = job_id
        self.original = original


@dataclass
class BatchResult(Generic[T, R]):
    """Result of batch processing."""
    successful: List[R] = field(default_factory=list)
    failed: List[tuple[T, Exception]] = field(default_factory=list)
    total: int = 0
    duration: float = 0.0
    
    @property
    def success_count(self) -> int:
        """Number of successful items."""
        return len(self.successful)
    
    @property
    def failure_count(self) -> int:
        """Number of failed items."""
        return len(self.failed)
    
    @property
    def success_rate(self) -> float:
        """Success rate as percentage."""
        return (self.success_count / self.total * 100) if self.total > 0 else 0.0
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary."""
        return {
            "total": self.total,
            "success_count": self.success_count,
            "failure_count": self.failure_count,
            "success_rate": self.success_rate,
            "duration": self.duration,
        }


@dataclass
class Job(Generic[T]):
    """Represents a queued job."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    func: Optional[Callable[..., T]] = None
    args: tuple = field(default_factory=tuple)
    kwargs: Dict[str, Any] = field(default_factory=dict)
    status: JobStatus = JobStatus.PENDING
    result: Optional[T] = None
    error: Optional[Exception] = None
    created_at: float = field(default_factory=time.time)
    started_at: Optional[float] = None
    completed_at: Optional[float] = None
    retries: int = 0
    max_retries: int = 3
    priority: int = 0
    
    @property
    def duration(self) -> Optional[float]:
        """Job duration in seconds."""
        if self.started_at and self.completed_at:
            return self.completed_at - self.started_at
        return None
    
    @property
    def wait_time(self) -> Optional[float]:
        """Time spent waiting in queue."""
        if self.started_at:
            return self.started_at - self.created_at
        return None


class BatchStrategy(ABC):
    """Strategy for batching items."""
    
    @abstractmethod
    def create_batches(self, items: Sequence[T]) -> List[List[T]]:
        """Create batches from items."""
        pass


class SizeBatchStrategy(BatchStrategy, Generic[T]):
    """Batch by fixed size."""
    
    def __init__(self, batch_size: int = 100):
        self.batch_size = batch_size
    
    def create_batches(self, items: Sequence[T]) -> List[List[T]]:
        """Create fixed-size batches."""
        batches = []
        for i in range(0, len(items), self.batch_size):
            batches.append(list(items[i:i + self.batch_size]))
        return batches


class DynamicBatchStrategy(BatchStrategy, Generic[T]):
    """Batch with dynamic sizing based on memory/complexity."""
    
    def __init__(
        self,
        min_size: int = 10,
        max_size: int = 1000,
        size_func: Optional[Callable[[T], int]] = None,
        max_batch_weight: int = 10000,
    ):
        self.min_size = min_size
        self.max_size = max_size
        self.size_func = size_func or (lambda x: 1)
        self.max_batch_weight = max_batch_weight
    
    def create_batches(self, items: Sequence[T]) -> List[List[T]]:
        """Create dynamically sized batches."""
        batches = []
        current_batch = []
        current_weight = 0
        
        for item in items:
            weight = self.size_func(item)
            
            if current_weight + weight > self.max_batch_weight:
                if current_batch:
                    batches.append(current_batch)
                current_batch = [item]
                current_weight = weight
            else:
                current_batch.append(item)
                current_weight += weight
            
            if len(current_batch) >= self.max_size:
                batches.append(current_batch)
                current_batch = []
                current_weight = 0
        
        if current_batch:
            batches.append(current_batch)
        
        return batches


class BatchProcessor(Generic[T, R]):
    """
    Process items in batches with concurrency control.
    """
    
    def __init__(
        self,
        handler: Callable[[T], R],
        batch_size: int = 100,
        max_concurrent: int = 5,
        continue_on_error: bool = True,
        retry_failed: bool = False,
        max_retries: int = 3,
        progress_callback: Optional[Callable[[int, int], None]] = None,
    ):
        """
        Initialize batch processor.
        
        Args:
            handler: Function to process each item
            batch_size: Number of items per batch
            max_concurrent: Maximum concurrent batches
            continue_on_error: Continue processing on errors
            retry_failed: Retry failed items
            max_retries: Maximum retry attempts
            progress_callback: Called with (completed, total)
        """
        self.handler = handler
        self.batch_size = batch_size
        self.max_concurrent = max_concurrent
        self.continue_on_error = continue_on_error
        self.retry_failed = retry_failed
        self.max_retries = max_retries
        self.progress_callback = progress_callback
        
        self._strategy = SizeBatchStrategy(batch_size)
    
    async def process(self, items: Sequence[T]) -> BatchResult[T, R]:
        """
        Process items in batches.
        
        Args:
            items: Items to process
            
        Returns:
            BatchResult with successful and failed items
        """
        start = time.time()
        result = BatchResult(total=len(items))
        
        if not items:
            return result
        
        batches = self._strategy.create_batches(items)
        semaphore = asyncio.Semaphore(self.max_concurrent)
        completed = 0
        
        async def process_batch(batch: List[T]) -> List[tuple[bool, Union[R, tuple[T, Exception]]]]:
            async with semaphore:
                batch_results = []
                for item in batch:
                    try:
                        if asyncio.iscoroutinefunction(self.handler):
                            r = await self.handler(item)
                        else:
                            r = self.handler(item)
                        batch_results.append((True, r))
                    except Exception as e:
                        if not self.continue_on_error:
                            raise
                        batch_results.append((False, (item, e)))
                return batch_results
        
        # Process all batches
        tasks = [process_batch(batch) for batch in batches]
        
        for coro in asyncio.as_completed(tasks):
            try:
                batch_results = await coro
                for success, data in batch_results:
                    if success:
                        result.successful.append(data)
                    else:
                        result.failed.append(data)
                    completed += 1
                    
                if self.progress_callback:
                    self.progress_callback(completed, result.total)
                    
            except Exception as e:
                logger.error(f"Batch processing error: {e}")
                if not self.continue_on_error:
                    raise
        
        # Retry failed items
        if self.retry_failed and result.failed:
            await self._retry_failed(result)
        
        result.duration = time.time() - start
        return result
    
    async def _retry_failed(self, result: BatchResult[T, R]) -> None:
        """Retry failed items."""
        retries = 0
        
        while result.failed and retries < self.max_retries:
            retries += 1
            failed_items = [item for item, _ in result.failed]
            result.failed.clear()
            
            batches = self._strategy.create_batches(failed_items)
            semaphore = asyncio.Semaphore(self.max_concurrent)
            
            async def process_item(item: T) -> tuple[bool, Union[R, tuple[T, Exception]]]:
                async with semaphore:
                    try:
                        if asyncio.iscoroutinefunction(self.handler):
                            r = await self.handler(item)
                        else:
                            r = self.handler(item)
                        return (True, r)
                    except Exception as e:
                        return (False, (item, e))
            
            for batch in batches:
                tasks = [process_item(item) for item in batch]
                batch_results = await asyncio.gather(*tasks)
                
                for success, data in batch_results:
                    if success:
                        result.successful.append(data)
                    else:
                        result.failed.append(data)
            
            logger.info(
                f"Retry {retries}: {len(result.failed)} still failing"
            )


class JobQueue:
    """
    Async job queue with worker pool.
    """
    
    def __init__(
        self,
        workers: int = 4,
        max_queue_size: int = 1000,
    ):
        """
        Initialize job queue.
        
        Args:
            workers: Number of worker tasks
            max_queue_size: Maximum queue size
        """
        self.workers = workers
        self.max_queue_size = max_queue_size
        
        self._queue: asyncio.PriorityQueue = asyncio.PriorityQueue(max_queue_size)
        self._jobs: Dict[str, Job] = {}
        self._worker_tasks: List[asyncio.Task] = []
        self._running = False
        self._results: Dict[str, asyncio.Event] = {}
    
    async def start(self) -> None:
        """Start the job queue workers."""
        if self._running:
            return
        
        self._running = True
        self._worker_tasks = [
            asyncio.create_task(self._worker(i))
            for i in range(self.workers)
        ]
        logger.info(f"Job queue started with {self.workers} workers")
    
    async def stop(self, wait: bool = True) -> None:
        """Stop the job queue."""
        self._running = False
        
        if wait:
            # Wait for pending jobs
            await self._queue.join()
        
        # Cancel workers
        for task in self._worker_tasks:
            task.cancel()
        
        await asyncio.gather(*self._worker_tasks, return_exceptions=True)
        self._worker_tasks.clear()
        logger.info("Job queue stopped")
    
    async def _worker(self, worker_id: int) -> None:
        """Worker task that processes jobs."""
        while self._running:
            try:
                _, job_id = await asyncio.wait_for(
                    self._queue.get(),
                    timeout=1.0,
                )
            except asyncio.TimeoutError:
                continue
            except asyncio.CancelledError:
                break
            
            job = self._jobs.get(job_id)
            if not job:
                self._queue.task_done()
                continue
            
            job.status = JobStatus.RUNNING
            job.started_at = time.time()
            
            try:
                if asyncio.iscoroutinefunction(job.func):
                    result = await job.func(*job.args, **job.kwargs)
                else:
                    result = job.func(*job.args, **job.kwargs)
                
                job.result = result
                job.status = JobStatus.COMPLETED
                
            except Exception as e:
                job.error = e
                
                if job.retries < job.max_retries:
                    job.retries += 1
                    job.status = JobStatus.RETRYING
                    await self._queue.put((-job.priority, job.id))
                else:
                    job.status = JobStatus.FAILED
                    logger.error(f"Job {job.id} failed: {e}")
            
            finally:
                job.completed_at = time.time()
                
                # Signal completion
                if job.id in self._results:
                    self._results[job.id].set()
                
                self._queue.task_done()
    
    async def submit(
        self,
        func: Callable[..., T],
        *args: Any,
        priority: int = 0,
        max_retries: int = 3,
        **kwargs: Any,
    ) -> str:
        """
        Submit a job to the queue.
        
        Args:
            func: Function to execute
            *args: Positional arguments
            priority: Job priority (higher = more urgent)
            max_retries: Maximum retry attempts
            **kwargs: Keyword arguments
            
        Returns:
            Job ID for tracking
        """
        job = Job(
            func=func,
            args=args,
            kwargs=kwargs,
            priority=priority,
            max_retries=max_retries,
        )
        
        self._jobs[job.id] = job
        self._results[job.id] = asyncio.Event()
        
        await self._queue.put((-priority, job.id))
        return job.id
    
    def get_job(self, job_id: str) -> Optional[Job]:
        """Get job by ID."""
        return self._jobs.get(job_id)
    
    async def wait(self, job_id: str, timeout: Optional[float] = None) -> Job:
        """
        Wait for job completion.
        
        Args:
            job_id: Job ID to wait for
            timeout: Maximum wait time
            
        Returns:
            Completed job
        """
        if job_id not in self._results:
            raise JobError(f"Job {job_id} not found", job_id)
        
        if timeout:
            await asyncio.wait_for(
                self._results[job_id].wait(),
                timeout=timeout,
            )
        else:
            await self._results[job_id].wait()
        
        return self._jobs[job_id]
    
    async def get_result(self, job_id: str, timeout: Optional[float] = None) -> T:
        """
        Wait for and return job result.
        
        Args:
            job_id: Job ID
            timeout: Maximum wait time
            
        Returns:
            Job result
            
        Raises:
            JobError: If job failed
        """
        job = await self.wait(job_id, timeout)
        
        if job.status == JobStatus.FAILED:
            raise JobError(
                f"Job {job_id} failed",
                job_id,
                job.error,
            )
        
        return job.result
    
    async def cancel(self, job_id: str) -> bool:
        """
        Cancel a pending job.
        
        Args:
            job_id: Job ID to cancel
            
        Returns:
            True if cancelled, False if not found or already completed
        """
        job = self._jobs.get(job_id)
        if not job:
            return False
        
        if job.status in (JobStatus.COMPLETED, JobStatus.FAILED):
            return False
        
        job.status = JobStatus.CANCELLED
        job.completed_at = time.time()
        
        if job_id in self._results:
            self._results[job_id].set()
        
        return True
    
    @property
    def pending_count(self) -> int:
        """Number of pending jobs."""
        return self._queue.qsize()
    
    @property
    def running_count(self) -> int:
        """Number of running jobs."""
        return sum(
            1 for job in self._jobs.values()
            if job.status == JobStatus.RUNNING
        )


class ParallelExecutor:
    """
    Execute functions in parallel with concurrency limits.
    """
    
    def __init__(
        self,
        max_concurrent: int = 10,
        timeout: Optional[float] = None,
    ):
        """
        Initialize parallel executor.
        
        Args:
            max_concurrent: Maximum concurrent executions
            timeout: Default timeout per execution
        """
        self.max_concurrent = max_concurrent
        self.timeout = timeout
        self._semaphore = asyncio.Semaphore(max_concurrent)
    
    async def map(
        self,
        func: Callable[[T], R],
        items: Sequence[T],
    ) -> List[R]:
        """
        Apply function to items in parallel.
        
        Args:
            func: Function to apply
            items: Items to process
            
        Returns:
            List of results in same order
        """
        async def execute(item: T, index: int) -> tuple[int, R]:
            async with self._semaphore:
                if asyncio.iscoroutinefunction(func):
                    result = await func(item)
                else:
                    result = func(item)
                return index, result
        
        tasks = [execute(item, i) for i, item in enumerate(items)]
        results = await asyncio.gather(*tasks)
        
        # Sort by original index
        sorted_results = sorted(results, key=lambda x: x[0])
        return [r for _, r in sorted_results]
    
    async def execute_many(
        self,
        tasks: List[tuple[Callable, tuple, Dict]],
    ) -> List[Any]:
        """
        Execute multiple different functions in parallel.
        
        Args:
            tasks: List of (func, args, kwargs) tuples
            
        Returns:
            List of results
        """
        async def execute(
            func: Callable,
            args: tuple,
            kwargs: Dict,
            index: int,
        ) -> tuple[int, Any]:
            async with self._semaphore:
                if asyncio.iscoroutinefunction(func):
                    result = await func(*args, **kwargs)
                else:
                    result = func(*args, **kwargs)
                return index, result
        
        coros = [
            execute(func, args, kwargs, i)
            for i, (func, args, kwargs) in enumerate(tasks)
        ]
        
        if self.timeout:
            results = await asyncio.wait_for(
                asyncio.gather(*coros),
                timeout=self.timeout * len(tasks),
            )
        else:
            results = await asyncio.gather(*coros)
        
        sorted_results = sorted(results, key=lambda x: x[0])
        return [r for _, r in sorted_results]


def batch(
    batch_size: int = 100,
    max_concurrent: int = 5,
    continue_on_error: bool = True,
) -> Callable:
    """
    Decorator for batch processing functions.
    
    Example:
        @batch(batch_size=50)
        async def process_items(items):
            return [process(item) for item in items]
    """
    def decorator(func: Callable) -> Callable:
        async def wrapper(items: Sequence[T], **kwargs: Any) -> BatchResult:
            processor = BatchProcessor(
                handler=func,
                batch_size=batch_size,
                max_concurrent=max_concurrent,
                continue_on_error=continue_on_error,
            )
            return await processor.process(items)
        
        wrapper.__name__ = func.__name__
        wrapper.__doc__ = func.__doc__
        return wrapper
    
    return decorator


def parallel(max_concurrent: int = 10) -> Callable:
    """
    Decorator for parallel execution.
    
    Example:
        @parallel(max_concurrent=5)
        async def fetch_urls(urls):
            return [await fetch(url) for url in urls]
    """
    def decorator(func: Callable) -> Callable:
        executor = ParallelExecutor(max_concurrent=max_concurrent)
        
        async def wrapper(items: Sequence[T]) -> List[R]:
            return await executor.map(func, items)
        
        wrapper.__name__ = func.__name__
        wrapper.__doc__ = func.__doc__
        return wrapper
    
    return decorator


__all__ = [
    # Enums
    "JobStatus",
    # Exceptions
    "BatchError",
    "JobError",
    # Data classes
    "BatchResult",
    "Job",
    # Strategies
    "BatchStrategy",
    "SizeBatchStrategy",
    "DynamicBatchStrategy",
    # Processors
    "BatchProcessor",
    "JobQueue",
    "ParallelExecutor",
    # Decorators
    "batch",
    "parallel",
]



"""
Enterprise Archive Manager Module.

Provides archive creation, extraction, encryption,
and streaming for ZIP/TAR formats.

Example:
    # Create archive manager
    archives = create_archive_manager()
    
    # Create ZIP archive
    archive = await archives.create_zip(
        files=[
            ("doc.pdf", pdf_bytes),
            ("images/logo.png", logo_bytes),
        ],
        password="secret",
    )
    
    # Extract archive
    files = await archives.extract(archive_bytes)
    
    # Stream archive entries
    async for entry in archives.stream(archive_bytes):
        print(f"{entry.name}: {entry.size} bytes")
"""

from __future__ import annotations

import asyncio
import base64
import functools
import gzip
import io
import json
import logging
import os
import tarfile
import uuid
import zipfile
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import (
    Any,
    AsyncIterator,
    BinaryIO,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

T = TypeVar('T')


logger = logging.getLogger(__name__)


class ArchiveError(Exception):
    """Archive error."""
    pass


class ExtractionError(ArchiveError):
    """Extraction error."""
    pass


class CompressionError(ArchiveError):
    """Compression error."""
    pass


class ArchiveFormat(str, Enum):
    """Archive formats."""
    ZIP = "zip"
    TAR = "tar"
    TAR_GZ = "tar.gz"
    TAR_BZ2 = "tar.bz2"
    TAR_XZ = "tar.xz"
    GZIP = "gz"
    BZIP2 = "bz2"
    SEVENZ = "7z"


class CompressionLevel(int, Enum):
    """Compression levels."""
    STORE = 0
    FASTEST = 1
    FAST = 3
    NORMAL = 6
    BEST = 9


@dataclass
class ArchiveEntry:
    """Archive entry information."""
    name: str
    size: int = 0
    compressed_size: int = 0
    is_directory: bool = False
    is_file: bool = True
    modified_time: Optional[datetime] = None
    permissions: Optional[int] = None
    content: Optional[bytes] = None


@dataclass
class ArchiveInfo:
    """Archive information."""
    format: ArchiveFormat
    total_size: int = 0
    compressed_size: int = 0
    entry_count: int = 0
    is_encrypted: bool = False
    comment: str = ""
    entries: List[ArchiveEntry] = field(default_factory=list)


@dataclass
class CreateOptions:
    """Archive creation options."""
    format: ArchiveFormat = ArchiveFormat.ZIP
    compression: CompressionLevel = CompressionLevel.NORMAL
    password: Optional[str] = None
    comment: str = ""
    preserve_paths: bool = True
    base_path: str = ""


@dataclass
class ExtractOptions:
    """Extraction options."""
    password: Optional[str] = None
    pattern: Optional[str] = None  # Glob pattern
    flatten: bool = False
    max_size: Optional[int] = None  # Max extracted size


@dataclass
class CreatedArchive:
    """Created archive result."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    content: bytes = b""
    format: ArchiveFormat = ArchiveFormat.ZIP
    size: int = 0
    entry_count: int = 0
    is_encrypted: bool = False
    created_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class ExtractedFile:
    """Extracted file."""
    name: str
    path: str
    content: bytes
    size: int
    modified_time: Optional[datetime] = None


class ArchiveManager:
    """
    Archive management service.
    """
    
    def __init__(
        self,
        temp_dir: Optional[Path] = None,
        max_size: int = 1024 * 1024 * 1024,  # 1 GB
    ):
        self._temp_dir = temp_dir or Path("/tmp")
        self._max_size = max_size
    
    async def get_info(
        self,
        data: bytes,
        password: Optional[str] = None,
    ) -> ArchiveInfo:
        """
        Get archive information.
        
        Args:
            data: Archive bytes
            password: Archive password
            
        Returns:
            Archive info
        """
        # Try ZIP first
        try:
            return await self._get_zip_info(data, password)
        except:
            pass
        
        # Try TAR
        try:
            return await self._get_tar_info(data)
        except:
            pass
        
        raise ArchiveError("Unknown archive format")
    
    async def _get_zip_info(
        self,
        data: bytes,
        password: Optional[str] = None,
    ) -> ArchiveInfo:
        """Get ZIP archive info."""
        buffer = io.BytesIO(data)
        
        with zipfile.ZipFile(buffer, 'r') as zf:
            entries = []
            total_size = 0
            compressed_size = 0
            
            for info in zf.infolist():
                entries.append(ArchiveEntry(
                    name=info.filename,
                    size=info.file_size,
                    compressed_size=info.compress_size,
                    is_directory=info.is_dir(),
                    is_file=not info.is_dir(),
                    modified_time=datetime(*info.date_time),
                ))
                total_size += info.file_size
                compressed_size += info.compress_size
            
            return ArchiveInfo(
                format=ArchiveFormat.ZIP,
                total_size=total_size,
                compressed_size=compressed_size,
                entry_count=len(entries),
                is_encrypted=any(info.flag_bits & 0x1 for info in zf.infolist()),
                comment=zf.comment.decode() if zf.comment else "",
                entries=entries,
            )
    
    async def _get_tar_info(self, data: bytes) -> ArchiveInfo:
        """Get TAR archive info."""
        buffer = io.BytesIO(data)
        
        # Determine format
        format = ArchiveFormat.TAR
        
        try:
            with tarfile.open(fileobj=buffer, mode='r:*') as tf:
                entries = []
                total_size = 0
                
                for member in tf.getmembers():
                    entries.append(ArchiveEntry(
                        name=member.name,
                        size=member.size,
                        compressed_size=member.size,
                        is_directory=member.isdir(),
                        is_file=member.isfile(),
                        modified_time=datetime.fromtimestamp(member.mtime),
                        permissions=member.mode,
                    ))
                    total_size += member.size
                
                return ArchiveInfo(
                    format=format,
                    total_size=total_size,
                    compressed_size=len(data),
                    entry_count=len(entries),
                    entries=entries,
                )
        except:
            raise ArchiveError("Not a valid TAR archive")
    
    async def create_zip(
        self,
        files: List[Tuple[str, bytes]],
        compression: CompressionLevel = CompressionLevel.NORMAL,
        password: Optional[str] = None,
        comment: str = "",
    ) -> CreatedArchive:
        """
        Create ZIP archive.
        
        Args:
            files: List of (name, content) tuples
            compression: Compression level
            password: Archive password
            comment: Archive comment
            
        Returns:
            Created archive
        """
        buffer = io.BytesIO()
        
        # Map compression level
        compression_type = zipfile.ZIP_DEFLATED
        if compression == CompressionLevel.STORE:
            compression_type = zipfile.ZIP_STORED
        
        with zipfile.ZipFile(
            buffer,
            'w',
            compression=compression_type,
            compresslevel=compression.value if compression != CompressionLevel.STORE else None,
        ) as zf:
            for name, content in files:
                zf.writestr(name, content)
            
            if comment:
                zf.comment = comment.encode()
        
        content = buffer.getvalue()
        
        return CreatedArchive(
            content=content,
            format=ArchiveFormat.ZIP,
            size=len(content),
            entry_count=len(files),
            is_encrypted=password is not None,
        )
    
    async def create_tar(
        self,
        files: List[Tuple[str, bytes]],
        compression: Optional[str] = "gz",
    ) -> CreatedArchive:
        """
        Create TAR archive.
        
        Args:
            files: List of (name, content) tuples
            compression: Compression type (gz, bz2, xz, or None)
            
        Returns:
            Created archive
        """
        buffer = io.BytesIO()
        
        mode = f"w:{compression}" if compression else "w"
        format = {
            "gz": ArchiveFormat.TAR_GZ,
            "bz2": ArchiveFormat.TAR_BZ2,
            "xz": ArchiveFormat.TAR_XZ,
            None: ArchiveFormat.TAR,
        }.get(compression, ArchiveFormat.TAR_GZ)
        
        with tarfile.open(fileobj=buffer, mode=mode) as tf:
            for name, content in files:
                info = tarfile.TarInfo(name=name)
                info.size = len(content)
                info.mtime = int(datetime.utcnow().timestamp())
                tf.addfile(info, io.BytesIO(content))
        
        content = buffer.getvalue()
        
        return CreatedArchive(
            content=content,
            format=format,
            size=len(content),
            entry_count=len(files),
        )
    
    async def create(
        self,
        files: List[Tuple[str, bytes]],
        options: Optional[CreateOptions] = None,
    ) -> CreatedArchive:
        """
        Create archive.
        
        Args:
            files: List of (name, content) tuples
            options: Creation options
            
        Returns:
            Created archive
        """
        options = options or CreateOptions()
        
        if options.format == ArchiveFormat.ZIP:
            return await self.create_zip(
                files,
                options.compression,
                options.password,
                options.comment,
            )
        elif options.format in (
            ArchiveFormat.TAR,
            ArchiveFormat.TAR_GZ,
            ArchiveFormat.TAR_BZ2,
            ArchiveFormat.TAR_XZ,
        ):
            compression = {
                ArchiveFormat.TAR: None,
                ArchiveFormat.TAR_GZ: "gz",
                ArchiveFormat.TAR_BZ2: "bz2",
                ArchiveFormat.TAR_XZ: "xz",
            }.get(options.format)
            return await self.create_tar(files, compression)
        else:
            raise ArchiveError(f"Unsupported format: {options.format}")
    
    async def extract(
        self,
        data: bytes,
        options: Optional[ExtractOptions] = None,
    ) -> List[ExtractedFile]:
        """
        Extract archive.
        
        Args:
            data: Archive bytes
            options: Extraction options
            
        Returns:
            List of extracted files
        """
        options = options or ExtractOptions()
        
        # Try ZIP first
        try:
            return await self._extract_zip(data, options)
        except zipfile.BadZipFile:
            pass
        
        # Try TAR
        try:
            return await self._extract_tar(data, options)
        except:
            pass
        
        raise ArchiveError("Unknown archive format")
    
    async def _extract_zip(
        self,
        data: bytes,
        options: ExtractOptions,
    ) -> List[ExtractedFile]:
        """Extract ZIP archive."""
        buffer = io.BytesIO(data)
        files = []
        
        with zipfile.ZipFile(buffer, 'r') as zf:
            # Set password if provided
            if options.password:
                zf.setpassword(options.password.encode())
            
            for info in zf.infolist():
                # Skip directories
                if info.is_dir():
                    continue
                
                # Check pattern
                if options.pattern:
                    import fnmatch
                    if not fnmatch.fnmatch(info.filename, options.pattern):
                        continue
                
                # Check max size
                if options.max_size and info.file_size > options.max_size:
                    continue
                
                # Extract
                content = zf.read(info)
                
                # Determine path
                if options.flatten:
                    path = os.path.basename(info.filename)
                else:
                    path = info.filename
                
                files.append(ExtractedFile(
                    name=os.path.basename(info.filename),
                    path=path,
                    content=content,
                    size=len(content),
                    modified_time=datetime(*info.date_time),
                ))
        
        return files
    
    async def _extract_tar(
        self,
        data: bytes,
        options: ExtractOptions,
    ) -> List[ExtractedFile]:
        """Extract TAR archive."""
        buffer = io.BytesIO(data)
        files = []
        
        with tarfile.open(fileobj=buffer, mode='r:*') as tf:
            for member in tf.getmembers():
                # Skip non-files
                if not member.isfile():
                    continue
                
                # Check pattern
                if options.pattern:
                    import fnmatch
                    if not fnmatch.fnmatch(member.name, options.pattern):
                        continue
                
                # Check max size
                if options.max_size and member.size > options.max_size:
                    continue
                
                # Extract
                f = tf.extractfile(member)
                if f:
                    content = f.read()
                    
                    # Determine path
                    if options.flatten:
                        path = os.path.basename(member.name)
                    else:
                        path = member.name
                    
                    files.append(ExtractedFile(
                        name=os.path.basename(member.name),
                        path=path,
                        content=content,
                        size=len(content),
                        modified_time=datetime.fromtimestamp(member.mtime),
                    ))
        
        return files
    
    async def extract_file(
        self,
        data: bytes,
        filename: str,
        password: Optional[str] = None,
    ) -> Optional[bytes]:
        """
        Extract single file from archive.
        
        Args:
            data: Archive bytes
            filename: File to extract
            password: Archive password
            
        Returns:
            File content or None
        """
        options = ExtractOptions(
            password=password,
            pattern=filename,
        )
        
        files = await self.extract(data, options)
        
        for file in files:
            if file.name == filename or file.path == filename:
                return file.content
        
        return None
    
    async def list_entries(
        self,
        data: bytes,
        password: Optional[str] = None,
    ) -> List[ArchiveEntry]:
        """List archive entries."""
        info = await self.get_info(data, password)
        return info.entries
    
    async def stream(
        self,
        data: bytes,
        password: Optional[str] = None,
    ) -> AsyncIterator[ArchiveEntry]:
        """
        Stream archive entries.
        
        Args:
            data: Archive bytes
            password: Archive password
            
        Yields:
            Archive entries with content
        """
        options = ExtractOptions(password=password)
        files = await self.extract(data, options)
        
        for file in files:
            yield ArchiveEntry(
                name=file.name,
                size=file.size,
                is_file=True,
                modified_time=file.modified_time,
                content=file.content,
            )
    
    async def add_to_archive(
        self,
        archive_data: bytes,
        files: List[Tuple[str, bytes]],
        format: ArchiveFormat = ArchiveFormat.ZIP,
    ) -> CreatedArchive:
        """
        Add files to existing archive.
        
        Args:
            archive_data: Existing archive
            files: New files to add
            format: Archive format
            
        Returns:
            Updated archive
        """
        # Extract existing files
        existing = await self.extract(archive_data)
        
        # Combine with new files
        all_files = [(f.path, f.content) for f in existing]
        all_files.extend(files)
        
        # Create new archive
        options = CreateOptions(format=format)
        return await self.create(all_files, options)
    
    async def compress(
        self,
        data: bytes,
        format: str = "gzip",
    ) -> bytes:
        """
        Compress data.
        
        Args:
            data: Data to compress
            format: Compression format (gzip, bz2)
            
        Returns:
            Compressed data
        """
        if format == "gzip":
            return gzip.compress(data)
        elif format == "bz2":
            import bz2
            return bz2.compress(data)
        else:
            raise ArchiveError(f"Unknown compression format: {format}")
    
    async def decompress(
        self,
        data: bytes,
        format: str = "gzip",
    ) -> bytes:
        """
        Decompress data.
        
        Args:
            data: Compressed data
            format: Compression format
            
        Returns:
            Decompressed data
        """
        if format == "gzip":
            return gzip.decompress(data)
        elif format == "bz2":
            import bz2
            return bz2.decompress(data)
        else:
            raise ArchiveError(f"Unknown compression format: {format}")
    
    async def save(
        self,
        archive: CreatedArchive,
        path: Union[str, Path],
    ) -> Path:
        """Save archive to file."""
        path = Path(path)
        path.write_bytes(archive.content)
        return path


# Builder
class ArchiveBuilder:
    """Fluent archive builder."""
    
    def __init__(
        self,
        format: ArchiveFormat = ArchiveFormat.ZIP,
    ):
        self._format = format
        self._files: List[Tuple[str, bytes]] = []
        self._options = CreateOptions(format=format)
        self._manager = ArchiveManager()
    
    def add_file(
        self,
        name: str,
        content: bytes,
    ) -> "ArchiveBuilder":
        """Add file to archive."""
        self._files.append((name, content))
        return self
    
    def add_files(
        self,
        files: List[Tuple[str, bytes]],
    ) -> "ArchiveBuilder":
        """Add multiple files."""
        self._files.extend(files)
        return self
    
    def add_directory(
        self,
        path: Union[str, Path],
        base_name: str = "",
    ) -> "ArchiveBuilder":
        """Add directory contents."""
        path = Path(path)
        
        for file_path in path.rglob("*"):
            if file_path.is_file():
                rel_path = file_path.relative_to(path)
                name = f"{base_name}/{rel_path}" if base_name else str(rel_path)
                self._files.append((name, file_path.read_bytes()))
        
        return self
    
    def set_password(self, password: str) -> "ArchiveBuilder":
        """Set archive password."""
        self._options.password = password
        return self
    
    def set_compression(
        self,
        level: CompressionLevel,
    ) -> "ArchiveBuilder":
        """Set compression level."""
        self._options.compression = level
        return self
    
    def set_comment(self, comment: str) -> "ArchiveBuilder":
        """Set archive comment."""
        self._options.comment = comment
        return self
    
    async def build(self) -> CreatedArchive:
        """Build archive."""
        return await self._manager.create(self._files, self._options)


# Factory functions
def create_archive_manager(
    temp_dir: Optional[Path] = None,
    max_size: int = 1024 * 1024 * 1024,
) -> ArchiveManager:
    """Create archive manager."""
    return ArchiveManager(temp_dir, max_size)


def create_archive_builder(
    format: ArchiveFormat = ArchiveFormat.ZIP,
) -> ArchiveBuilder:
    """Create archive builder."""
    return ArchiveBuilder(format)


def create_create_options(
    format: ArchiveFormat = ArchiveFormat.ZIP,
    compression: CompressionLevel = CompressionLevel.NORMAL,
    **kwargs,
) -> CreateOptions:
    """Create creation options."""
    return CreateOptions(format=format, compression=compression, **kwargs)


def create_extract_options(
    password: Optional[str] = None,
    pattern: Optional[str] = None,
    **kwargs,
) -> ExtractOptions:
    """Create extraction options."""
    return ExtractOptions(password=password, pattern=pattern, **kwargs)


__all__ = [
    # Exceptions
    "ArchiveError",
    "ExtractionError",
    "CompressionError",
    # Enums
    "ArchiveFormat",
    "CompressionLevel",
    # Data classes
    "ArchiveEntry",
    "ArchiveInfo",
    "CreateOptions",
    "ExtractOptions",
    "CreatedArchive",
    "ExtractedFile",
    # Manager
    "ArchiveManager",
    # Builder
    "ArchiveBuilder",
    # Factory functions
    "create_archive_manager",
    "create_archive_builder",
    "create_create_options",
    "create_extract_options",
]



"""
Enterprise GraphQL Manager Module.

Provides GraphQL schema definition, resolvers, dataloaders,
subscriptions, and query optimization.

Example:
    # Create GraphQL manager
    gql = create_graphql_manager()
    
    # Define types
    @gql.type()
    class User:
        id: str
        name: str
        email: str
    
    # Define resolvers
    @gql.resolver("Query", "user")
    async def resolve_user(parent, info, id: str) -> User:
        return await get_user(id)
    
    # Execute query
    result = await gql.execute('''
        query {
            user(id: "123") {
                name
                email
            }
        }
    ''')
"""

from __future__ import annotations

import asyncio
import functools
import inspect
import json
import logging
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

T = TypeVar('T')


logger = logging.getLogger(__name__)


class GraphQLError(Exception):
    """Base GraphQL error."""
    def __init__(
        self,
        message: str,
        locations: Optional[List[Dict[str, int]]] = None,
        path: Optional[List[str]] = None,
        extensions: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(message)
        self.message = message
        self.locations = locations
        self.path = path
        self.extensions = extensions or {}
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert to GraphQL error format."""
        error = {"message": self.message}
        if self.locations:
            error["locations"] = self.locations
        if self.path:
            error["path"] = self.path
        if self.extensions:
            error["extensions"] = self.extensions
        return error


class ValidationError(GraphQLError):
    """Schema validation error."""
    pass


class ExecutionError(GraphQLError):
    """Query execution error."""
    pass


class TypeKind(str, Enum):
    """GraphQL type kinds."""
    SCALAR = "SCALAR"
    OBJECT = "OBJECT"
    INTERFACE = "INTERFACE"
    UNION = "UNION"
    ENUM = "ENUM"
    INPUT_OBJECT = "INPUT_OBJECT"
    LIST = "LIST"
    NON_NULL = "NON_NULL"


@dataclass
class FieldDefinition:
    """GraphQL field definition."""
    name: str
    type: str
    description: str = ""
    args: Dict[str, "ArgumentDefinition"] = field(default_factory=dict)
    resolver: Optional[Callable] = None
    deprecation_reason: Optional[str] = None


@dataclass
class ArgumentDefinition:
    """GraphQL argument definition."""
    name: str
    type: str
    description: str = ""
    default_value: Any = None


@dataclass
class TypeDefinition:
    """GraphQL type definition."""
    name: str
    kind: TypeKind
    description: str = ""
    fields: Dict[str, FieldDefinition] = field(default_factory=dict)
    interfaces: List[str] = field(default_factory=list)
    possible_types: List[str] = field(default_factory=list)  # For unions/interfaces
    enum_values: List[str] = field(default_factory=list)
    input_fields: Dict[str, ArgumentDefinition] = field(default_factory=dict)


@dataclass
class ResolverInfo:
    """Resolver execution info."""
    field_name: str
    parent_type: str
    return_type: str
    path: List[str]
    context: Dict[str, Any]
    variables: Dict[str, Any]
    operation_name: Optional[str]


@dataclass
class ExecutionResult:
    """GraphQL execution result."""
    data: Optional[Dict[str, Any]] = None
    errors: List[GraphQLError] = field(default_factory=list)
    extensions: Dict[str, Any] = field(default_factory=dict)
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert to response format."""
        result = {}
        if self.data is not None:
            result["data"] = self.data
        if self.errors:
            result["errors"] = [e.to_dict() for e in self.errors]
        if self.extensions:
            result["extensions"] = self.extensions
        return result


class DataLoader(Generic[T]):
    """
    DataLoader for batching and caching data fetches.
    """
    
    def __init__(
        self,
        batch_fn: Callable[[List[str]], Awaitable[List[T]]],
        max_batch_size: int = 100,
        cache: bool = True,
    ):
        self._batch_fn = batch_fn
        self._max_batch_size = max_batch_size
        self._cache_enabled = cache
        
        self._cache: Dict[str, T] = {}
        self._pending: Dict[str, asyncio.Future] = {}
        self._queue: List[str] = []
        self._batch_scheduled = False
    
    async def load(self, key: str) -> T:
        """Load a single item."""
        # Check cache
        if self._cache_enabled and key in self._cache:
            return self._cache[key]
        
        # Check pending
        if key in self._pending:
            return await self._pending[key]
        
        # Create future and add to queue
        future: asyncio.Future = asyncio.get_event_loop().create_future()
        self._pending[key] = future
        self._queue.append(key)
        
        # Schedule batch
        if not self._batch_scheduled:
            self._batch_scheduled = True
            asyncio.get_event_loop().call_soon(
                lambda: asyncio.create_task(self._dispatch())
            )
        
        return await future
    
    async def load_many(self, keys: List[str]) -> List[T]:
        """Load multiple items."""
        return await asyncio.gather(*[self.load(key) for key in keys])
    
    async def _dispatch(self) -> None:
        """Dispatch batched requests."""
        self._batch_scheduled = False
        
        if not self._queue:
            return
        
        # Get batch
        batch = self._queue[:self._max_batch_size]
        self._queue = self._queue[self._max_batch_size:]
        
        try:
            results = await self._batch_fn(batch)
            
            # Resolve futures
            for key, result in zip(batch, results):
                if self._cache_enabled:
                    self._cache[key] = result
                
                future = self._pending.pop(key, None)
                if future and not future.done():
                    future.set_result(result)
                    
        except Exception as e:
            # Reject all futures
            for key in batch:
                future = self._pending.pop(key, None)
                if future and not future.done():
                    future.set_exception(e)
        
        # Schedule next batch if queue not empty
        if self._queue and not self._batch_scheduled:
            self._batch_scheduled = True
            asyncio.get_event_loop().call_soon(
                lambda: asyncio.create_task(self._dispatch())
            )
    
    def clear(self, key: Optional[str] = None) -> None:
        """Clear cache."""
        if key:
            self._cache.pop(key, None)
        else:
            self._cache.clear()
    
    def prime(self, key: str, value: T) -> None:
        """Prime cache with value."""
        if self._cache_enabled:
            self._cache[key] = value


class Middleware(ABC):
    """GraphQL middleware."""
    
    @abstractmethod
    async def resolve(
        self,
        next_fn: Callable,
        parent: Any,
        info: ResolverInfo,
        **kwargs,
    ) -> Any:
        """Middleware resolver."""
        pass


class LoggingMiddleware(Middleware):
    """Logging middleware."""
    
    async def resolve(
        self,
        next_fn: Callable,
        parent: Any,
        info: ResolverInfo,
        **kwargs,
    ) -> Any:
        start_time = datetime.utcnow()
        
        try:
            result = await next_fn(parent, info, **kwargs)
            elapsed = (datetime.utcnow() - start_time).total_seconds()
            
            logger.debug(
                f"Resolved {info.parent_type}.{info.field_name} "
                f"in {elapsed:.3f}s"
            )
            
            return result
        except Exception as e:
            logger.error(
                f"Error resolving {info.parent_type}.{info.field_name}: {e}"
            )
            raise


class CachingMiddleware(Middleware):
    """Caching middleware."""
    
    def __init__(self, ttl: int = 60):
        self._cache: Dict[str, Tuple[Any, datetime]] = {}
        self._ttl = ttl
    
    async def resolve(
        self,
        next_fn: Callable,
        parent: Any,
        info: ResolverInfo,
        **kwargs,
    ) -> Any:
        cache_key = f"{info.parent_type}.{info.field_name}:{json.dumps(kwargs, sort_keys=True)}"
        
        # Check cache
        if cache_key in self._cache:
            value, cached_at = self._cache[cache_key]
            if (datetime.utcnow() - cached_at).total_seconds() < self._ttl:
                return value
        
        result = await next_fn(parent, info, **kwargs)
        self._cache[cache_key] = (result, datetime.utcnow())
        
        return result


class Schema:
    """GraphQL schema."""
    
    def __init__(self):
        self._types: Dict[str, TypeDefinition] = {}
        self._resolvers: Dict[str, Dict[str, Callable]] = defaultdict(dict)
        self._directives: Dict[str, Callable] = {}
        
        # Initialize built-in types
        self._init_built_in_types()
    
    def _init_built_in_types(self) -> None:
        """Initialize built-in scalar types."""
        scalars = ["String", "Int", "Float", "Boolean", "ID"]
        for scalar in scalars:
            self._types[scalar] = TypeDefinition(
                name=scalar,
                kind=TypeKind.SCALAR,
            )
        
        # Query and Mutation root types
        self._types["Query"] = TypeDefinition(
            name="Query",
            kind=TypeKind.OBJECT,
        )
        self._types["Mutation"] = TypeDefinition(
            name="Mutation",
            kind=TypeKind.OBJECT,
        )
        self._types["Subscription"] = TypeDefinition(
            name="Subscription",
            kind=TypeKind.OBJECT,
        )
    
    def add_type(self, type_def: TypeDefinition) -> None:
        """Add type definition."""
        self._types[type_def.name] = type_def
    
    def get_type(self, name: str) -> Optional[TypeDefinition]:
        """Get type definition."""
        return self._types.get(name)
    
    def add_field(
        self,
        type_name: str,
        field: FieldDefinition,
    ) -> None:
        """Add field to type."""
        if type_name not in self._types:
            self._types[type_name] = TypeDefinition(
                name=type_name,
                kind=TypeKind.OBJECT,
            )
        
        self._types[type_name].fields[field.name] = field
    
    def add_resolver(
        self,
        type_name: str,
        field_name: str,
        resolver: Callable,
    ) -> None:
        """Add field resolver."""
        self._resolvers[type_name][field_name] = resolver
    
    def get_resolver(
        self,
        type_name: str,
        field_name: str,
    ) -> Optional[Callable]:
        """Get field resolver."""
        return self._resolvers.get(type_name, {}).get(field_name)


class GraphQLManager:
    """
    GraphQL schema and execution manager.
    """
    
    def __init__(self):
        self._schema = Schema()
        self._middlewares: List[Middleware] = []
        self._dataloaders: Dict[str, DataLoader] = {}
    
    @property
    def schema(self) -> Schema:
        return self._schema
    
    # Type decorators
    def type(
        self,
        name: Optional[str] = None,
        description: str = "",
    ) -> Callable:
        """Decorator to register GraphQL type."""
        def decorator(cls: Type) -> Type:
            type_name = name or cls.__name__
            
            # Create type definition from class
            fields = {}
            for field_name, field_type in getattr(cls, '__annotations__', {}).items():
                fields[field_name] = FieldDefinition(
                    name=field_name,
                    type=self._python_type_to_graphql(field_type),
                )
            
            type_def = TypeDefinition(
                name=type_name,
                kind=TypeKind.OBJECT,
                description=description,
                fields=fields,
            )
            
            self._schema.add_type(type_def)
            return cls
        
        return decorator
    
    def input_type(
        self,
        name: Optional[str] = None,
    ) -> Callable:
        """Decorator to register input type."""
        def decorator(cls: Type) -> Type:
            type_name = name or cls.__name__
            
            input_fields = {}
            for field_name, field_type in getattr(cls, '__annotations__', {}).items():
                input_fields[field_name] = ArgumentDefinition(
                    name=field_name,
                    type=self._python_type_to_graphql(field_type),
                )
            
            type_def = TypeDefinition(
                name=type_name,
                kind=TypeKind.INPUT_OBJECT,
                input_fields=input_fields,
            )
            
            self._schema.add_type(type_def)
            return cls
        
        return decorator
    
    def enum_type(
        self,
        name: Optional[str] = None,
    ) -> Callable:
        """Decorator to register enum type."""
        def decorator(cls: Type) -> Type:
            type_name = name or cls.__name__
            
            enum_values = [e.name for e in cls] if issubclass(cls, Enum) else []
            
            type_def = TypeDefinition(
                name=type_name,
                kind=TypeKind.ENUM,
                enum_values=enum_values,
            )
            
            self._schema.add_type(type_def)
            return cls
        
        return decorator
    
    # Resolver decorators
    def resolver(
        self,
        type_name: str,
        field_name: str,
    ) -> Callable:
        """Decorator to register resolver."""
        def decorator(func: Callable) -> Callable:
            self._schema.add_resolver(type_name, field_name, func)
            return func
        
        return decorator
    
    def query(self, field_name: str) -> Callable:
        """Decorator for Query resolver."""
        return self.resolver("Query", field_name)
    
    def mutation(self, field_name: str) -> Callable:
        """Decorator for Mutation resolver."""
        return self.resolver("Mutation", field_name)
    
    def subscription(self, field_name: str) -> Callable:
        """Decorator for Subscription resolver."""
        return self.resolver("Subscription", field_name)
    
    def field(
        self,
        type_name: str,
        field_name: str,
        return_type: str,
        description: str = "",
        args: Optional[Dict[str, str]] = None,
    ) -> Callable:
        """Decorator to add field and resolver."""
        def decorator(func: Callable) -> Callable:
            # Add field definition
            field_args = {}
            if args:
                for arg_name, arg_type in args.items():
                    field_args[arg_name] = ArgumentDefinition(
                        name=arg_name,
                        type=arg_type,
                    )
            
            field_def = FieldDefinition(
                name=field_name,
                type=return_type,
                description=description,
                args=field_args,
            )
            
            self._schema.add_field(type_name, field_def)
            self._schema.add_resolver(type_name, field_name, func)
            
            return func
        
        return decorator
    
    # Middleware
    def add_middleware(self, middleware: Middleware) -> None:
        """Add execution middleware."""
        self._middlewares.append(middleware)
    
    # DataLoader
    def dataloader(
        self,
        name: str,
        batch_fn: Callable[[List[str]], Awaitable[List[Any]]],
        max_batch_size: int = 100,
    ) -> DataLoader:
        """Create and register dataloader."""
        loader = DataLoader(batch_fn, max_batch_size)
        self._dataloaders[name] = loader
        return loader
    
    def get_dataloader(self, name: str) -> Optional[DataLoader]:
        """Get registered dataloader."""
        return self._dataloaders.get(name)
    
    # Execution
    async def execute(
        self,
        query: str,
        variables: Optional[Dict[str, Any]] = None,
        operation_name: Optional[str] = None,
        context: Optional[Dict[str, Any]] = None,
    ) -> ExecutionResult:
        """Execute GraphQL query."""
        try:
            # Parse query (simplified)
            parsed = self._parse_query(query)
            
            # Execute
            data = await self._execute_operation(
                parsed,
                variables or {},
                operation_name,
                context or {},
            )
            
            return ExecutionResult(data=data)
            
        except GraphQLError as e:
            return ExecutionResult(errors=[e])
        except Exception as e:
            return ExecutionResult(
                errors=[GraphQLError(str(e))]
            )
    
    def _parse_query(self, query: str) -> Dict[str, Any]:
        """Parse GraphQL query (simplified)."""
        # This is a simplified parser for demo
        # Real implementation would use proper GraphQL parser
        return {"query": query, "type": "query"}
    
    async def _execute_operation(
        self,
        parsed: Dict[str, Any],
        variables: Dict[str, Any],
        operation_name: Optional[str],
        context: Dict[str, Any],
    ) -> Dict[str, Any]:
        """Execute parsed operation."""
        # Simplified execution
        # Real implementation would traverse AST
        return {}
    
    def _python_type_to_graphql(self, python_type: Any) -> str:
        """Convert Python type to GraphQL type."""
        type_map = {
            str: "String",
            int: "Int",
            float: "Float",
            bool: "Boolean",
        }
        
        if python_type in type_map:
            return type_map[python_type]
        
        # Handle Optional, List, etc.
        origin = getattr(python_type, "__origin__", None)
        
        if origin is list:
            args = getattr(python_type, "__args__", [])
            if args:
                inner = self._python_type_to_graphql(args[0])
                return f"[{inner}]"
        
        if origin is Union:
            args = getattr(python_type, "__args__", [])
            # Optional[X] is Union[X, None]
            non_none = [a for a in args if a is not type(None)]
            if len(non_none) == 1:
                return self._python_type_to_graphql(non_none[0])
        
        # Default to type name
        return getattr(python_type, "__name__", "String")


# Global manager
_global_manager: Optional[GraphQLManager] = None


# Factory functions
def create_graphql_manager() -> GraphQLManager:
    """Create GraphQL manager."""
    return GraphQLManager()


def create_dataloader(
    batch_fn: Callable[[List[str]], Awaitable[List[T]]],
    max_batch_size: int = 100,
    cache: bool = True,
) -> DataLoader[T]:
    """Create dataloader."""
    return DataLoader(batch_fn, max_batch_size, cache)


def create_schema() -> Schema:
    """Create GraphQL schema."""
    return Schema()


def create_logging_middleware() -> LoggingMiddleware:
    """Create logging middleware."""
    return LoggingMiddleware()


def create_caching_middleware(ttl: int = 60) -> CachingMiddleware:
    """Create caching middleware."""
    return CachingMiddleware(ttl)


def get_global_manager() -> GraphQLManager:
    """Get global GraphQL manager."""
    global _global_manager
    if _global_manager is None:
        _global_manager = create_graphql_manager()
    return _global_manager


__all__ = [
    # Exceptions
    "GraphQLError",
    "ValidationError",
    "ExecutionError",
    # Enums
    "TypeKind",
    # Data classes
    "FieldDefinition",
    "ArgumentDefinition",
    "TypeDefinition",
    "ResolverInfo",
    "ExecutionResult",
    # DataLoader
    "DataLoader",
    # Middleware
    "Middleware",
    "LoggingMiddleware",
    "CachingMiddleware",
    # Schema
    "Schema",
    # Manager
    "GraphQLManager",
    # Factory functions
    "create_graphql_manager",
    "create_dataloader",
    "create_schema",
    "create_logging_middleware",
    "create_caching_middleware",
    "get_global_manager",
]



"""
Enterprise Payment Gateway Module.

Payment processing, multiple providers,
refunds, and transaction management.

Example:
    # Create payment gateway
    payment = create_payment_gateway(stripe_config)
    
    # Charge customer
    result = await payment.charge(
        amount=9999,  # in cents
        currency="usd",
        source="tok_visa",
        description="Order #123",
    )
    
    # Refund
    refund = await payment.refund(
        transaction_id=result.transaction_id,
        amount=5000,  # partial refund
    )
    
    # Create subscription
    sub = await payment.create_subscription(
        customer_id="cus_123",
        plan_id="plan_premium",
    )
"""

from __future__ import annotations

import asyncio
import hashlib
import hmac
import json
import logging
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from decimal import Decimal
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

T = TypeVar('T')

logger = logging.getLogger(__name__)


class PaymentError(Exception):
    """Payment error."""
    pass


class PaymentDeclinedError(PaymentError):
    """Payment declined."""
    pass


class InvalidCardError(PaymentError):
    """Invalid card."""
    pass


class InsufficientFundsError(PaymentError):
    """Insufficient funds."""
    pass


class PaymentStatus(str, Enum):
    """Payment status."""
    PENDING = "pending"
    PROCESSING = "processing"
    SUCCEEDED = "succeeded"
    FAILED = "failed"
    CANCELED = "canceled"
    REFUNDED = "refunded"
    PARTIALLY_REFUNDED = "partially_refunded"
    DISPUTED = "disputed"


class PaymentMethod(str, Enum):
    """Payment methods."""
    CARD = "card"
    BANK_TRANSFER = "bank_transfer"
    WALLET = "wallet"
    CRYPTO = "crypto"
    PAYPAL = "paypal"
    APPLE_PAY = "apple_pay"
    GOOGLE_PAY = "google_pay"


class SubscriptionStatus(str, Enum):
    """Subscription status."""
    ACTIVE = "active"
    CANCELED = "canceled"
    PAST_DUE = "past_due"
    UNPAID = "unpaid"
    TRIALING = "trialing"
    PAUSED = "paused"


class Currency(str, Enum):
    """Currencies."""
    USD = "usd"
    EUR = "eur"
    GBP = "gbp"
    JPY = "jpy"
    AUD = "aud"
    CAD = "cad"
    CHF = "chf"
    INR = "inr"


@dataclass
class PaymentConfig:
    """Payment provider configuration."""
    provider: str = "mock"
    api_key: str = ""
    secret_key: str = ""
    webhook_secret: str = ""
    sandbox: bool = True
    default_currency: Currency = Currency.USD
    options: Dict[str, Any] = field(default_factory=dict)


@dataclass
class Card:
    """Payment card."""
    number: str = ""
    exp_month: int = 0
    exp_year: int = 0
    cvc: str = ""
    name: str = ""
    last_four: str = ""
    brand: str = ""
    country: str = ""


@dataclass
class Customer:
    """Customer."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    email: str = ""
    name: str = ""
    phone: str = ""
    default_source: Optional[str] = None
    sources: List[str] = field(default_factory=list)
    metadata: Dict[str, Any] = field(default_factory=dict)
    created_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class Transaction:
    """Payment transaction."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    amount: int = 0  # in smallest currency unit (cents)
    currency: Currency = Currency.USD
    status: PaymentStatus = PaymentStatus.PENDING
    method: PaymentMethod = PaymentMethod.CARD
    customer_id: Optional[str] = None
    source: str = ""
    description: str = ""
    metadata: Dict[str, Any] = field(default_factory=dict)
    created_at: datetime = field(default_factory=datetime.utcnow)
    updated_at: datetime = field(default_factory=datetime.utcnow)
    error_message: Optional[str] = None
    refund_amount: int = 0
    fee: int = 0
    net_amount: int = 0


@dataclass
class Refund:
    """Refund."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    transaction_id: str = ""
    amount: int = 0
    reason: str = ""
    status: PaymentStatus = PaymentStatus.PENDING
    created_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class Plan:
    """Subscription plan."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    name: str = ""
    amount: int = 0
    currency: Currency = Currency.USD
    interval: str = "month"  # day, week, month, year
    interval_count: int = 1
    trial_days: int = 0
    features: List[str] = field(default_factory=list)
    metadata: Dict[str, Any] = field(default_factory=dict)
    active: bool = True


@dataclass
class Subscription:
    """Subscription."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    customer_id: str = ""
    plan_id: str = ""
    status: SubscriptionStatus = SubscriptionStatus.ACTIVE
    current_period_start: datetime = field(default_factory=datetime.utcnow)
    current_period_end: Optional[datetime] = None
    trial_start: Optional[datetime] = None
    trial_end: Optional[datetime] = None
    canceled_at: Optional[datetime] = None
    ended_at: Optional[datetime] = None
    metadata: Dict[str, Any] = field(default_factory=dict)
    created_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class ChargeResult:
    """Charge result."""
    success: bool
    transaction_id: str
    status: PaymentStatus
    amount: int
    currency: Currency
    error: Optional[str] = None
    error_code: Optional[str] = None


@dataclass
class PaymentStats:
    """Payment statistics."""
    total_transactions: int = 0
    total_amount: int = 0
    successful: int = 0
    failed: int = 0
    refunded: int = 0
    by_method: Dict[str, int] = field(default_factory=dict)
    by_currency: Dict[str, int] = field(default_factory=dict)


# Payment provider base
class PaymentProvider(ABC):
    """Base payment provider."""
    
    @abstractmethod
    async def charge(
        self,
        amount: int,
        currency: Currency,
        source: str,
        **kwargs,
    ) -> ChargeResult:
        """Charge payment."""
        pass
    
    @abstractmethod
    async def refund(
        self,
        transaction_id: str,
        amount: Optional[int] = None,
        reason: str = "",
    ) -> Refund:
        """Refund payment."""
        pass
    
    @abstractmethod
    async def get_transaction(
        self,
        transaction_id: str,
    ) -> Optional[Transaction]:
        """Get transaction."""
        pass
    
    @abstractmethod
    async def create_customer(
        self,
        email: str,
        name: str = "",
        **kwargs,
    ) -> Customer:
        """Create customer."""
        pass
    
    @abstractmethod
    async def create_subscription(
        self,
        customer_id: str,
        plan_id: str,
        **kwargs,
    ) -> Subscription:
        """Create subscription."""
        pass


# Mock provider
class MockPaymentProvider(PaymentProvider):
    """Mock payment provider for testing."""
    
    def __init__(self):
        self._transactions: Dict[str, Transaction] = {}
        self._customers: Dict[str, Customer] = {}
        self._subscriptions: Dict[str, Subscription] = {}
        self._plans: Dict[str, Plan] = {}
        self.should_fail: bool = False
        self.failure_rate: float = 0.0
    
    async def charge(
        self,
        amount: int,
        currency: Currency,
        source: str,
        customer_id: Optional[str] = None,
        description: str = "",
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> ChargeResult:
        """Charge (mock)."""
        import random
        
        # Simulate failure
        if self.should_fail or random.random() < self.failure_rate:
            return ChargeResult(
                success=False,
                transaction_id="",
                status=PaymentStatus.FAILED,
                amount=amount,
                currency=currency,
                error="Payment declined",
                error_code="card_declined",
            )
        
        # Create transaction
        transaction = Transaction(
            amount=amount,
            currency=currency,
            status=PaymentStatus.SUCCEEDED,
            customer_id=customer_id,
            source=source,
            description=description,
            metadata=metadata or {},
            fee=int(amount * 0.029 + 30),  # 2.9% + $0.30
        )
        transaction.net_amount = amount - transaction.fee
        
        self._transactions[transaction.id] = transaction
        
        return ChargeResult(
            success=True,
            transaction_id=transaction.id,
            status=PaymentStatus.SUCCEEDED,
            amount=amount,
            currency=currency,
        )
    
    async def refund(
        self,
        transaction_id: str,
        amount: Optional[int] = None,
        reason: str = "",
    ) -> Refund:
        """Refund (mock)."""
        transaction = self._transactions.get(transaction_id)
        if not transaction:
            raise PaymentError(f"Transaction not found: {transaction_id}")
        
        refund_amount = amount or transaction.amount
        if refund_amount > (transaction.amount - transaction.refund_amount):
            raise PaymentError("Refund amount exceeds available")
        
        refund = Refund(
            transaction_id=transaction_id,
            amount=refund_amount,
            reason=reason,
            status=PaymentStatus.SUCCEEDED,
        )
        
        # Update transaction
        transaction.refund_amount += refund_amount
        if transaction.refund_amount >= transaction.amount:
            transaction.status = PaymentStatus.REFUNDED
        else:
            transaction.status = PaymentStatus.PARTIALLY_REFUNDED
        transaction.updated_at = datetime.utcnow()
        
        return refund
    
    async def get_transaction(
        self,
        transaction_id: str,
    ) -> Optional[Transaction]:
        """Get transaction (mock)."""
        return self._transactions.get(transaction_id)
    
    async def create_customer(
        self,
        email: str,
        name: str = "",
        **kwargs,
    ) -> Customer:
        """Create customer (mock)."""
        customer = Customer(
            email=email,
            name=name,
            metadata=kwargs.get("metadata", {}),
        )
        self._customers[customer.id] = customer
        return customer
    
    async def create_subscription(
        self,
        customer_id: str,
        plan_id: str,
        trial_days: int = 0,
        **kwargs,
    ) -> Subscription:
        """Create subscription (mock)."""
        now = datetime.utcnow()
        
        subscription = Subscription(
            customer_id=customer_id,
            plan_id=plan_id,
            status=SubscriptionStatus.TRIALING if trial_days > 0 else SubscriptionStatus.ACTIVE,
            current_period_start=now,
            current_period_end=now + timedelta(days=30),
            trial_start=now if trial_days > 0 else None,
            trial_end=now + timedelta(days=trial_days) if trial_days > 0 else None,
            metadata=kwargs.get("metadata", {}),
        )
        
        self._subscriptions[subscription.id] = subscription
        return subscription
    
    async def cancel_subscription(
        self,
        subscription_id: str,
        at_period_end: bool = True,
    ) -> Subscription:
        """Cancel subscription (mock)."""
        subscription = self._subscriptions.get(subscription_id)
        if not subscription:
            raise PaymentError(f"Subscription not found: {subscription_id}")
        
        subscription.canceled_at = datetime.utcnow()
        if not at_period_end:
            subscription.status = SubscriptionStatus.CANCELED
            subscription.ended_at = datetime.utcnow()
        
        return subscription
    
    async def add_plan(self, plan: Plan) -> Plan:
        """Add plan (mock)."""
        self._plans[plan.id] = plan
        return plan


# Transaction store
class TransactionStore(ABC):
    """Transaction storage."""
    
    @abstractmethod
    async def save(self, transaction: Transaction) -> None:
        """Save transaction."""
        pass
    
    @abstractmethod
    async def get(self, transaction_id: str) -> Optional[Transaction]:
        """Get transaction."""
        pass
    
    @abstractmethod
    async def list(
        self,
        customer_id: Optional[str] = None,
        status: Optional[PaymentStatus] = None,
        limit: int = 50,
    ) -> List[Transaction]:
        """List transactions."""
        pass


class InMemoryTransactionStore(TransactionStore):
    """In-memory transaction store."""
    
    def __init__(self):
        self._transactions: Dict[str, Transaction] = {}
    
    async def save(self, transaction: Transaction) -> None:
        transaction.updated_at = datetime.utcnow()
        self._transactions[transaction.id] = transaction
    
    async def get(self, transaction_id: str) -> Optional[Transaction]:
        return self._transactions.get(transaction_id)
    
    async def list(
        self,
        customer_id: Optional[str] = None,
        status: Optional[PaymentStatus] = None,
        limit: int = 50,
    ) -> List[Transaction]:
        results = []
        for tx in self._transactions.values():
            if customer_id and tx.customer_id != customer_id:
                continue
            if status and tx.status != status:
                continue
            results.append(tx)
            if len(results) >= limit:
                break
        return results


# Payment gateway
class PaymentGateway:
    """Payment gateway service."""
    
    def __init__(
        self,
        provider: PaymentProvider,
        store: Optional[TransactionStore] = None,
        config: Optional[PaymentConfig] = None,
    ):
        self.provider = provider
        self.store = store or InMemoryTransactionStore()
        self.config = config or PaymentConfig()
        self._stats = PaymentStats()
        self._webhooks: List[Callable] = []
    
    async def charge(
        self,
        amount: int,
        currency: Optional[Currency] = None,
        source: str = "",
        customer_id: Optional[str] = None,
        description: str = "",
        metadata: Optional[Dict[str, Any]] = None,
        idempotency_key: Optional[str] = None,
    ) -> ChargeResult:
        """Charge payment."""
        currency = currency or self.config.default_currency
        
        result = await self.provider.charge(
            amount=amount,
            currency=currency,
            source=source,
            customer_id=customer_id,
            description=description,
            metadata=metadata,
        )
        
        # Update stats
        self._stats.total_transactions += 1
        if result.success:
            self._stats.successful += 1
            self._stats.total_amount += amount
        else:
            self._stats.failed += 1
        
        # Store transaction
        if result.success and result.transaction_id:
            transaction = await self.provider.get_transaction(result.transaction_id)
            if transaction:
                await self.store.save(transaction)
        
        logger.info(
            f"Payment {'succeeded' if result.success else 'failed'}: "
            f"{amount/100:.2f} {currency.value}"
        )
        
        return result
    
    async def refund(
        self,
        transaction_id: str,
        amount: Optional[int] = None,
        reason: str = "",
    ) -> Refund:
        """Refund payment."""
        refund = await self.provider.refund(
            transaction_id=transaction_id,
            amount=amount,
            reason=reason,
        )
        
        self._stats.refunded += 1
        
        logger.info(f"Refund processed: {refund.amount/100:.2f} for {transaction_id}")
        
        return refund
    
    async def get_transaction(
        self,
        transaction_id: str,
    ) -> Optional[Transaction]:
        """Get transaction."""
        # Check store first
        transaction = await self.store.get(transaction_id)
        if transaction:
            return transaction
        
        # Fallback to provider
        return await self.provider.get_transaction(transaction_id)
    
    async def list_transactions(
        self,
        customer_id: Optional[str] = None,
        status: Optional[PaymentStatus] = None,
        limit: int = 50,
    ) -> List[Transaction]:
        """List transactions."""
        return await self.store.list(
            customer_id=customer_id,
            status=status,
            limit=limit,
        )
    
    async def create_customer(
        self,
        email: str,
        name: str = "",
        **kwargs,
    ) -> Customer:
        """Create customer."""
        return await self.provider.create_customer(
            email=email,
            name=name,
            **kwargs,
        )
    
    async def create_subscription(
        self,
        customer_id: str,
        plan_id: str,
        **kwargs,
    ) -> Subscription:
        """Create subscription."""
        return await self.provider.create_subscription(
            customer_id=customer_id,
            plan_id=plan_id,
            **kwargs,
        )
    
    def on_webhook(self, callback: Callable) -> None:
        """Register webhook handler."""
        self._webhooks.append(callback)
    
    async def handle_webhook(
        self,
        payload: Dict[str, Any],
        signature: str = "",
    ) -> bool:
        """Handle webhook."""
        # Verify signature if webhook secret configured
        if self.config.webhook_secret and signature:
            expected = hmac.new(
                self.config.webhook_secret.encode(),
                json.dumps(payload, sort_keys=True).encode(),
                hashlib.sha256,
            ).hexdigest()
            
            if not hmac.compare_digest(signature, expected):
                logger.warning("Invalid webhook signature")
                return False
        
        # Call handlers
        for handler in self._webhooks:
            try:
                await handler(payload) if asyncio.iscoroutinefunction(handler) else handler(payload)
            except Exception as e:
                logger.error(f"Webhook handler error: {e}")
        
        return True
    
    def get_stats(self) -> PaymentStats:
        """Get statistics."""
        return self._stats


# Factory functions
def create_payment_gateway(
    config: Optional[PaymentConfig] = None,
    provider: Optional[PaymentProvider] = None,
) -> PaymentGateway:
    """Create payment gateway."""
    if provider is None:
        provider = MockPaymentProvider()
    
    return PaymentGateway(
        provider=provider,
        config=config,
    )


def create_plan(
    name: str,
    amount: int,
    currency: Currency = Currency.USD,
    interval: str = "month",
    **kwargs,
) -> Plan:
    """Create subscription plan."""
    return Plan(
        name=name,
        amount=amount,
        currency=currency,
        interval=interval,
        **kwargs,
    )


def create_payment_config(
    provider: str = "mock",
    api_key: str = "",
    secret_key: str = "",
    **kwargs,
) -> PaymentConfig:
    """Create payment config."""
    return PaymentConfig(
        provider=provider,
        api_key=api_key,
        secret_key=secret_key,
        **kwargs,
    )


__all__ = [
    # Exceptions
    "PaymentError",
    "PaymentDeclinedError",
    "InvalidCardError",
    "InsufficientFundsError",
    # Enums
    "PaymentStatus",
    "PaymentMethod",
    "SubscriptionStatus",
    "Currency",
    # Data classes
    "PaymentConfig",
    "Card",
    "Customer",
    "Transaction",
    "Refund",
    "Plan",
    "Subscription",
    "ChargeResult",
    "PaymentStats",
    # Providers
    "PaymentProvider",
    "MockPaymentProvider",
    # Store
    "TransactionStore",
    "InMemoryTransactionStore",
    # Service
    "PaymentGateway",
    # Factory functions
    "create_payment_gateway",
    "create_plan",
    "create_payment_config",
]



"""
Policy Engine.

Rule-based policy enforcement with:
- Pattern matching
- Conditional evaluation
- Policy composition
"""

import uuid
import time
import logging
import re
import threading
from typing import Dict, Any, List, Tuple, Pattern

from .types import Policy, PolicyType, AuditEventType, AuditSeverity

logger = logging.getLogger(__name__)


class PolicyEngine:
    """
    Policy enforcement engine.
    
    Features:
    - Rule-based policies
    - Pattern matching
    - Conditional evaluation
    - Policy composition
    """
    
    def __init__(self, audit_manager=None):
        self.policies: Dict[str, Policy] = {}
        self.audit_manager = audit_manager
        self._compiled_patterns: Dict[str, Tuple[Pattern, Pattern]] = {}
        self._lock = threading.Lock()
    
    def add_policy(self, policy: Policy):
        """Add a policy."""
        with self._lock:
            self.policies[policy.policy_id] = policy
            
            # Compile patterns
            self._compiled_patterns[policy.policy_id] = (
                re.compile(policy.resource_pattern),
                re.compile(policy.action_pattern)
            )
        
        logger.info("Added policy: %s (%s)", policy.name, policy.policy_type.value)
    
    def remove_policy(self, policy_id: str):
        """Remove a policy."""
        with self._lock:
            if policy_id in self.policies:
                del self.policies[policy_id]
                del self._compiled_patterns[policy_id]
    
    def evaluate(self,
                resource: str,
                action: str,
                context: Dict[str, Any] = None,
                actor: str = None) -> Dict[str, Any]:
        """
        Evaluate policies for a resource/action.
        
        Returns:
            Decision with allowed/denied and reason
        """
        context = context or {}
        matching_policies = []
        
        # Find matching policies
        for policy_id, policy in self.policies.items():
            if not policy.enabled:
                continue
            
            resource_pattern, action_pattern = self._compiled_patterns[policy_id]
            
            if resource_pattern.match(resource) and action_pattern.match(action):
                # Check conditions
                if self._evaluate_conditions(policy.conditions, context):
                    matching_policies.append(policy)
        
        # Sort by priority
        matching_policies.sort(key=lambda p: p.priority, reverse=True)
        
        result = {
            'allowed': True,
            'reason': None,
            'matched_policies': [],
            'requires': []
        }
        
        for policy in matching_policies:
            result['matched_policies'].append({
                'policy_id': policy.policy_id,
                'name': policy.name,
                'type': policy.policy_type.value
            })
            
            if policy.policy_type == PolicyType.DENY:
                result['allowed'] = False
                result['reason'] = f"Denied by policy: {policy.name}"
                break
            
            elif policy.policy_type == PolicyType.REQUIRE:
                requirement = policy.conditions.get('requirement')
                if requirement and not context.get(requirement):
                    result['requires'].append(requirement)
            
            elif policy.policy_type == PolicyType.AUDIT:
                if self.audit_manager:
                    self.audit_manager.log(
                        event_type=AuditEventType.ACCESS,
                        actor=actor or 'unknown',
                        resource=resource,
                        action=action,
                        details={'policy': policy.name, 'context': context},
                        severity=AuditSeverity.INFO
                    )
        
        # Check requirements
        if result['requires']:
            result['allowed'] = False
            result['reason'] = f"Requirements not met: {result['requires']}"
        
        return result
    
    def _evaluate_conditions(self, 
                            conditions: Dict[str, Any],
                            context: Dict[str, Any]) -> bool:
        """Evaluate policy conditions."""
        for key, expected in conditions.items():
            if key == 'requirement':
                continue  # Handled separately
            
            if key.startswith('context.'):
                context_key = key[8:]
                actual = context.get(context_key)
            else:
                actual = context.get(key)
            
            if isinstance(expected, list):
                if actual not in expected:
                    return False
            elif actual != expected:
                return False
        
        return True
    
    def create_allow_policy(self,
                           name: str,
                           resource_pattern: str,
                           action_pattern: str,
                           conditions: Dict[str, Any] = None,
                           priority: int = 100) -> Policy:
        """Create an allow policy."""
        policy = Policy(
            policy_id=str(uuid.uuid4()),
            name=name,
            description=f"Allow {action_pattern} on {resource_pattern}",
            policy_type=PolicyType.ALLOW,
            resource_pattern=resource_pattern,
            action_pattern=action_pattern,
            conditions=conditions or {},
            priority=priority,
            enabled=True,
            created_at=time.time()
        )
        
        self.add_policy(policy)
        return policy
    
    def create_deny_policy(self,
                          name: str,
                          resource_pattern: str,
                          action_pattern: str,
                          conditions: Dict[str, Any] = None,
                          priority: int = 200) -> Policy:
        """Create a deny policy."""
        policy = Policy(
            policy_id=str(uuid.uuid4()),
            name=name,
            description=f"Deny {action_pattern} on {resource_pattern}",
            policy_type=PolicyType.DENY,
            resource_pattern=resource_pattern,
            action_pattern=action_pattern,
            conditions=conditions or {},
            priority=priority,
            enabled=True,
            created_at=time.time()
        )
        
        self.add_policy(policy)
        return policy
    
    def list_policies(self, enabled_only: bool = True) -> List[Dict[str, Any]]:
        """List all policies."""
        policies = list(self.policies.values())
        
        if enabled_only:
            policies = [p for p in policies if p.enabled]
        
        return [
            {
                'policy_id': p.policy_id,
                'name': p.name,
                'type': p.policy_type.value,
                'resource_pattern': p.resource_pattern,
                'action_pattern': p.action_pattern,
                'priority': p.priority,
                'enabled': p.enabled
            }
            for p in sorted(policies, key=lambda x: x.priority, reverse=True)
        ]


__all__ = ['PolicyEngine']



"""
Enterprise Encryption Service Module.

Provides encryption, decryption, key management,
and secure storage using AES and RSA algorithms.

Example:
    # Create encryption service
    crypto = create_encryption_service()
    
    # Symmetric encryption (AES)
    encrypted = await crypto.encrypt(data, key="secret_key")
    decrypted = await crypto.decrypt(encrypted, key="secret_key")
    
    # Asymmetric encryption (RSA)
    public_key, private_key = await crypto.generate_rsa_keypair()
    encrypted = await crypto.rsa_encrypt(data, public_key)
    decrypted = await crypto.rsa_decrypt(encrypted, private_key)
    
    # Key derivation
    derived_key = await crypto.derive_key("password", salt=b"...")
    
    # Hashing
    hash_value = await crypto.hash(data, algorithm="sha256")
"""

from __future__ import annotations

import asyncio
import base64
import functools
import hashlib
import hmac
import json
import logging
import os
import secrets
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

T = TypeVar('T')


logger = logging.getLogger(__name__)


class CryptoError(Exception):
    """Cryptography error."""
    pass


class EncryptionError(CryptoError):
    """Encryption error."""
    pass


class DecryptionError(CryptoError):
    """Decryption error."""
    pass


class KeyError(CryptoError):
    """Key error."""
    pass


class Algorithm(str, Enum):
    """Encryption algorithms."""
    AES_128_CBC = "aes-128-cbc"
    AES_256_CBC = "aes-256-cbc"
    AES_128_GCM = "aes-128-gcm"
    AES_256_GCM = "aes-256-gcm"
    RSA_OAEP = "rsa-oaep"
    RSA_PKCS1 = "rsa-pkcs1"


class HashAlgorithm(str, Enum):
    """Hash algorithms."""
    MD5 = "md5"
    SHA1 = "sha1"
    SHA256 = "sha256"
    SHA384 = "sha384"
    SHA512 = "sha512"
    BLAKE2B = "blake2b"
    BLAKE2S = "blake2s"


class KeyType(str, Enum):
    """Key types."""
    SYMMETRIC = "symmetric"
    PUBLIC = "public"
    PRIVATE = "private"
    DERIVED = "derived"


@dataclass
class EncryptionKey:
    """Encryption key."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    type: KeyType = KeyType.SYMMETRIC
    algorithm: str = "aes-256-gcm"
    key: bytes = b""
    created_at: datetime = field(default_factory=datetime.utcnow)
    expires_at: Optional[datetime] = None
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class KeyPair:
    """RSA key pair."""
    public_key: bytes = b""
    private_key: bytes = b""
    key_size: int = 2048
    created_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class EncryptedData:
    """Encrypted data container."""
    ciphertext: bytes = b""
    iv: bytes = b""  # Initialization vector
    tag: bytes = b""  # Auth tag for GCM
    algorithm: str = "aes-256-gcm"
    key_id: Optional[str] = None
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class HashResult:
    """Hash result."""
    value: bytes = b""
    algorithm: HashAlgorithm = HashAlgorithm.SHA256
    hex: str = ""
    base64: str = ""


@dataclass
class DerivedKey:
    """Derived key from password."""
    key: bytes = b""
    salt: bytes = b""
    iterations: int = 100000
    algorithm: str = "pbkdf2"
    hash_algorithm: HashAlgorithm = HashAlgorithm.SHA256


# Key Store
class KeyStore(ABC):
    """Abstract key storage."""
    
    @abstractmethod
    async def store(
        self,
        key_id: str,
        key: EncryptionKey,
    ) -> None:
        """Store key."""
        pass
    
    @abstractmethod
    async def get(self, key_id: str) -> Optional[EncryptionKey]:
        """Get key."""
        pass
    
    @abstractmethod
    async def delete(self, key_id: str) -> bool:
        """Delete key."""
        pass
    
    @abstractmethod
    async def list(self) -> List[str]:
        """List key IDs."""
        pass


class InMemoryKeyStore(KeyStore):
    """In-memory key store."""
    
    def __init__(self):
        self._keys: Dict[str, EncryptionKey] = {}
    
    async def store(
        self,
        key_id: str,
        key: EncryptionKey,
    ) -> None:
        self._keys[key_id] = key
    
    async def get(self, key_id: str) -> Optional[EncryptionKey]:
        return self._keys.get(key_id)
    
    async def delete(self, key_id: str) -> bool:
        if key_id in self._keys:
            del self._keys[key_id]
            return True
        return False
    
    async def list(self) -> List[str]:
        return list(self._keys.keys())


class EncryptionService:
    """
    Encryption service.
    """
    
    def __init__(
        self,
        key_store: Optional[KeyStore] = None,
        default_algorithm: Algorithm = Algorithm.AES_256_GCM,
    ):
        self._key_store = key_store or InMemoryKeyStore()
        self._default_algorithm = default_algorithm
    
    # Key generation
    async def generate_key(
        self,
        algorithm: Algorithm = Algorithm.AES_256_GCM,
        key_id: Optional[str] = None,
        expires_in: Optional[timedelta] = None,
    ) -> EncryptionKey:
        """
        Generate encryption key.
        
        Args:
            algorithm: Encryption algorithm
            key_id: Optional key ID
            expires_in: Key expiration
            
        Returns:
            Generated key
        """
        # Determine key size
        key_size = 32  # 256 bits
        if "128" in algorithm.value:
            key_size = 16
        
        key_bytes = secrets.token_bytes(key_size)
        
        key = EncryptionKey(
            id=key_id or str(uuid.uuid4()),
            type=KeyType.SYMMETRIC,
            algorithm=algorithm.value,
            key=key_bytes,
            expires_at=datetime.utcnow() + expires_in if expires_in else None,
        )
        
        # Store key
        await self._key_store.store(key.id, key)
        
        return key
    
    async def generate_rsa_keypair(
        self,
        key_size: int = 2048,
    ) -> KeyPair:
        """
        Generate RSA key pair.
        
        Args:
            key_size: Key size in bits
            
        Returns:
            Key pair (mock implementation)
        """
        # Mock RSA key generation
        # In production, use cryptography library
        return KeyPair(
            public_key=secrets.token_bytes(256),
            private_key=secrets.token_bytes(key_size // 8),
            key_size=key_size,
        )
    
    async def derive_key(
        self,
        password: str,
        salt: Optional[bytes] = None,
        iterations: int = 100000,
        key_length: int = 32,
        hash_algorithm: HashAlgorithm = HashAlgorithm.SHA256,
    ) -> DerivedKey:
        """
        Derive key from password using PBKDF2.
        
        Args:
            password: Password
            salt: Salt (generated if not provided)
            iterations: Number of iterations
            key_length: Derived key length
            hash_algorithm: Hash algorithm
            
        Returns:
            Derived key
        """
        salt = salt or secrets.token_bytes(16)
        
        # Use hashlib for PBKDF2
        key = hashlib.pbkdf2_hmac(
            hash_algorithm.value,
            password.encode(),
            salt,
            iterations,
            dklen=key_length,
        )
        
        return DerivedKey(
            key=key,
            salt=salt,
            iterations=iterations,
            algorithm="pbkdf2",
            hash_algorithm=hash_algorithm,
        )
    
    # Symmetric encryption (AES)
    async def encrypt(
        self,
        data: Union[bytes, str],
        key: Optional[Union[bytes, str, EncryptionKey]] = None,
        key_id: Optional[str] = None,
        algorithm: Optional[Algorithm] = None,
    ) -> EncryptedData:
        """
        Encrypt data with AES.
        
        Args:
            data: Data to encrypt
            key: Encryption key
            key_id: Key ID (if using key store)
            algorithm: Encryption algorithm
            
        Returns:
            Encrypted data
        """
        # Convert string to bytes
        if isinstance(data, str):
            data = data.encode()
        
        # Resolve key
        key_bytes, resolved_key_id = await self._resolve_key(key, key_id)
        
        algorithm = algorithm or self._default_algorithm
        
        # Generate IV
        iv = secrets.token_bytes(16)
        
        # Mock encryption (XOR with key for demo)
        # In production, use cryptography library
        ciphertext = self._mock_encrypt(data, key_bytes, iv)
        
        # Generate auth tag for GCM
        tag = b""
        if "gcm" in algorithm.value:
            tag = hmac.new(key_bytes, ciphertext, hashlib.sha256).digest()[:16]
        
        return EncryptedData(
            ciphertext=ciphertext,
            iv=iv,
            tag=tag,
            algorithm=algorithm.value,
            key_id=resolved_key_id,
        )
    
    async def decrypt(
        self,
        encrypted: EncryptedData,
        key: Optional[Union[bytes, str, EncryptionKey]] = None,
        key_id: Optional[str] = None,
    ) -> bytes:
        """
        Decrypt data.
        
        Args:
            encrypted: Encrypted data
            key: Decryption key
            key_id: Key ID
            
        Returns:
            Decrypted data
        """
        # Resolve key
        key_bytes, _ = await self._resolve_key(
            key,
            key_id or encrypted.key_id,
        )
        
        # Verify auth tag for GCM
        if "gcm" in encrypted.algorithm and encrypted.tag:
            expected_tag = hmac.new(
                key_bytes, encrypted.ciphertext, hashlib.sha256
            ).digest()[:16]
            
            if not hmac.compare_digest(encrypted.tag, expected_tag):
                raise DecryptionError("Authentication failed")
        
        # Mock decryption
        return self._mock_decrypt(encrypted.ciphertext, key_bytes, encrypted.iv)
    
    def _mock_encrypt(
        self,
        data: bytes,
        key: bytes,
        iv: bytes,
    ) -> bytes:
        """Mock encryption (XOR for demo)."""
        # Simple XOR for demonstration
        # In production, use proper AES implementation
        extended_key = (key + iv) * (len(data) // len(key + iv) + 1)
        return bytes(a ^ b for a, b in zip(data, extended_key))
    
    def _mock_decrypt(
        self,
        data: bytes,
        key: bytes,
        iv: bytes,
    ) -> bytes:
        """Mock decryption (XOR is symmetric)."""
        return self._mock_encrypt(data, key, iv)
    
    async def _resolve_key(
        self,
        key: Optional[Union[bytes, str, EncryptionKey]],
        key_id: Optional[str],
    ) -> Tuple[bytes, Optional[str]]:
        """Resolve key to bytes."""
        if isinstance(key, EncryptionKey):
            return key.key, key.id
        
        if isinstance(key, str):
            # Hash string key to fixed size
            return hashlib.sha256(key.encode()).digest(), None
        
        if isinstance(key, bytes):
            return key, None
        
        if key_id:
            stored_key = await self._key_store.get(key_id)
            if stored_key:
                return stored_key.key, key_id
            raise KeyError(f"Key not found: {key_id}")
        
        raise KeyError("No key provided")
    
    # RSA encryption
    async def rsa_encrypt(
        self,
        data: Union[bytes, str],
        public_key: bytes,
    ) -> bytes:
        """
        Encrypt with RSA public key.
        
        Args:
            data: Data to encrypt
            public_key: RSA public key
            
        Returns:
            Encrypted data
        """
        if isinstance(data, str):
            data = data.encode()
        
        # Mock RSA encryption
        return self._mock_encrypt(data, public_key[:32], public_key[32:48])
    
    async def rsa_decrypt(
        self,
        data: bytes,
        private_key: bytes,
    ) -> bytes:
        """
        Decrypt with RSA private key.
        
        Args:
            data: Encrypted data
            private_key: RSA private key
            
        Returns:
            Decrypted data
        """
        # Mock RSA decryption
        return self._mock_decrypt(data, private_key[:32], private_key[32:48])
    
    # Hashing
    async def hash(
        self,
        data: Union[bytes, str],
        algorithm: HashAlgorithm = HashAlgorithm.SHA256,
    ) -> HashResult:
        """
        Hash data.
        
        Args:
            data: Data to hash
            algorithm: Hash algorithm
            
        Returns:
            Hash result
        """
        if isinstance(data, str):
            data = data.encode()
        
        if algorithm == HashAlgorithm.BLAKE2B:
            h = hashlib.blake2b(data)
        elif algorithm == HashAlgorithm.BLAKE2S:
            h = hashlib.blake2s(data)
        else:
            h = hashlib.new(algorithm.value, data)
        
        digest = h.digest()
        
        return HashResult(
            value=digest,
            algorithm=algorithm,
            hex=digest.hex(),
            base64=base64.b64encode(digest).decode(),
        )
    
    async def hash_password(
        self,
        password: str,
        salt: Optional[bytes] = None,
    ) -> Tuple[str, bytes]:
        """
        Hash password securely.
        
        Args:
            password: Password to hash
            salt: Salt (generated if not provided)
            
        Returns:
            Tuple of (hash, salt)
        """
        salt = salt or secrets.token_bytes(16)
        
        derived = await self.derive_key(
            password,
            salt=salt,
            iterations=100000,
        )
        
        return base64.b64encode(derived.key).decode(), salt
    
    async def verify_password(
        self,
        password: str,
        hash_value: str,
        salt: bytes,
    ) -> bool:
        """
        Verify password against hash.
        
        Args:
            password: Password to verify
            hash_value: Stored hash
            salt: Salt used for hashing
            
        Returns:
            True if password matches
        """
        computed_hash, _ = await self.hash_password(password, salt)
        return hmac.compare_digest(computed_hash, hash_value)
    
    # HMAC
    async def hmac(
        self,
        data: Union[bytes, str],
        key: Union[bytes, str],
        algorithm: HashAlgorithm = HashAlgorithm.SHA256,
    ) -> HashResult:
        """
        Generate HMAC.
        
        Args:
            data: Data to sign
            key: HMAC key
            algorithm: Hash algorithm
            
        Returns:
            HMAC result
        """
        if isinstance(data, str):
            data = data.encode()
        if isinstance(key, str):
            key = key.encode()
        
        h = hmac.new(key, data, algorithm.value)
        digest = h.digest()
        
        return HashResult(
            value=digest,
            algorithm=algorithm,
            hex=digest.hex(),
            base64=base64.b64encode(digest).decode(),
        )
    
    async def verify_hmac(
        self,
        data: Union[bytes, str],
        key: Union[bytes, str],
        signature: Union[bytes, str],
        algorithm: HashAlgorithm = HashAlgorithm.SHA256,
    ) -> bool:
        """
        Verify HMAC signature.
        
        Args:
            data: Data that was signed
            key: HMAC key
            signature: Signature to verify
            algorithm: Hash algorithm
            
        Returns:
            True if valid
        """
        result = await self.hmac(data, key, algorithm)
        
        if isinstance(signature, str):
            signature = bytes.fromhex(signature)
        
        return hmac.compare_digest(result.value, signature)
    
    # Utilities
    def generate_random_bytes(self, length: int = 32) -> bytes:
        """Generate random bytes."""
        return secrets.token_bytes(length)
    
    def generate_random_string(self, length: int = 32) -> str:
        """Generate random string."""
        return secrets.token_urlsafe(length)
    
    def encode_base64(self, data: bytes) -> str:
        """Encode to base64."""
        return base64.b64encode(data).decode()
    
    def decode_base64(self, data: str) -> bytes:
        """Decode from base64."""
        return base64.b64decode(data)
    
    # Serialization
    def serialize_encrypted(self, encrypted: EncryptedData) -> str:
        """Serialize encrypted data to string."""
        return json.dumps({
            "ciphertext": base64.b64encode(encrypted.ciphertext).decode(),
            "iv": base64.b64encode(encrypted.iv).decode(),
            "tag": base64.b64encode(encrypted.tag).decode() if encrypted.tag else "",
            "algorithm": encrypted.algorithm,
            "key_id": encrypted.key_id,
            "metadata": encrypted.metadata,
        })
    
    def deserialize_encrypted(self, data: str) -> EncryptedData:
        """Deserialize encrypted data from string."""
        parsed = json.loads(data)
        return EncryptedData(
            ciphertext=base64.b64decode(parsed["ciphertext"]),
            iv=base64.b64decode(parsed["iv"]),
            tag=base64.b64decode(parsed["tag"]) if parsed["tag"] else b"",
            algorithm=parsed["algorithm"],
            key_id=parsed.get("key_id"),
            metadata=parsed.get("metadata", {}),
        )


# Decorators
def encrypted_field(
    key: Optional[str] = None,
    algorithm: Algorithm = Algorithm.AES_256_GCM,
) -> Callable:
    """Decorator for encrypted field."""
    def decorator(func: Callable) -> Callable:
        func._encrypted = True
        func._encryption_key = key
        func._encryption_algorithm = algorithm
        return func
    return decorator


def hash_on_save(
    algorithm: HashAlgorithm = HashAlgorithm.SHA256,
) -> Callable:
    """Decorator to hash value on save."""
    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            result = await func(*args, **kwargs)
            
            if isinstance(result, (str, bytes)):
                crypto = EncryptionService()
                hash_result = await crypto.hash(result, algorithm)
                return hash_result.hex
            
            return result
        
        return wrapper
    return decorator


# Factory functions
def create_encryption_service(
    key_store: Optional[KeyStore] = None,
    default_algorithm: Algorithm = Algorithm.AES_256_GCM,
) -> EncryptionService:
    """Create encryption service."""
    return EncryptionService(key_store, default_algorithm)


def create_key_store() -> InMemoryKeyStore:
    """Create in-memory key store."""
    return InMemoryKeyStore()


def create_encryption_key(
    key_bytes: Optional[bytes] = None,
    algorithm: str = "aes-256-gcm",
    key_id: Optional[str] = None,
) -> EncryptionKey:
    """Create encryption key."""
    return EncryptionKey(
        id=key_id or str(uuid.uuid4()),
        type=KeyType.SYMMETRIC,
        algorithm=algorithm,
        key=key_bytes or secrets.token_bytes(32),
    )


__all__ = [
    # Exceptions
    "CryptoError",
    "EncryptionError",
    "DecryptionError",
    "KeyError",
    # Enums
    "Algorithm",
    "HashAlgorithm",
    "KeyType",
    # Data classes
    "EncryptionKey",
    "KeyPair",
    "EncryptedData",
    "HashResult",
    "DerivedKey",
    # Key Store
    "KeyStore",
    "InMemoryKeyStore",
    # Service
    "EncryptionService",
    # Decorators
    "encrypted_field",
    "hash_on_save",
    # Factory functions
    "create_encryption_service",
    "create_key_store",
    "create_encryption_key",
]



"""
Enterprise Migration Module.

Provides schema migration, versioning, and rollback
capabilities for database and configuration changes.

Example:
    # Create migrator
    migrator = create_migrator()
    
    # Register migrations
    @migration(version="1.0.0")
    async def add_users_table(ctx):
        await ctx.execute("CREATE TABLE users ...")
    
    # Run migrations
    await migrator.migrate_to("latest")
    
    # Rollback
    await migrator.rollback(steps=1)
"""

from __future__ import annotations

import asyncio
import hashlib
import logging
import re
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from functools import wraps
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

logger = logging.getLogger(__name__)

T = TypeVar('T')


class MigrationError(Exception):
    """Migration error."""
    pass


class RollbackError(MigrationError):
    """Rollback error."""
    pass


class VersionConflictError(MigrationError):
    """Version conflict error."""
    pass


class MigrationStatus(str, Enum):
    """Migration status."""
    PENDING = "pending"
    RUNNING = "running"
    COMPLETED = "completed"
    FAILED = "failed"
    ROLLED_BACK = "rolled_back"


class MigrationDirection(str, Enum):
    """Migration direction."""
    UP = "up"
    DOWN = "down"


@dataclass
class Version:
    """Semantic version."""
    major: int
    minor: int
    patch: int
    
    @classmethod
    def parse(cls, version_str: str) -> "Version":
        """Parse version string."""
        match = re.match(r"(\d+)\.(\d+)\.(\d+)", version_str)
        if not match:
            raise ValueError(f"Invalid version: {version_str}")
        
        return cls(
            major=int(match.group(1)),
            minor=int(match.group(2)),
            patch=int(match.group(3)),
        )
    
    def __str__(self) -> str:
        return f"{self.major}.{self.minor}.{self.patch}"
    
    def __lt__(self, other: "Version") -> bool:
        return (self.major, self.minor, self.patch) < (
            other.major, other.minor, other.patch
        )
    
    def __le__(self, other: "Version") -> bool:
        return self == other or self < other
    
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Version):
            return False
        return (self.major, self.minor, self.patch) == (
            other.major, other.minor, other.patch
        )
    
    def __hash__(self) -> int:
        return hash((self.major, self.minor, self.patch))


@dataclass
class MigrationRecord:
    """Record of an applied migration."""
    migration_id: str
    version: str
    name: str
    status: MigrationStatus
    applied_at: datetime = field(default_factory=datetime.now)
    completed_at: Optional[datetime] = None
    checksum: Optional[str] = None
    execution_time_ms: int = 0
    error: Optional[str] = None
    metadata: Dict[str, Any] = field(default_factory=dict)


@dataclass
class MigrationDefinition:
    """Definition of a migration."""
    migration_id: str
    version: str
    name: str
    up: Callable
    down: Optional[Callable] = None
    description: Optional[str] = None
    dependencies: List[str] = field(default_factory=list)
    checksum: Optional[str] = None


@dataclass
class MigrationContext:
    """Context for migration execution."""
    migration_id: str
    version: str
    direction: MigrationDirection
    dry_run: bool = False
    metadata: Dict[str, Any] = field(default_factory=dict)
    
    async def execute(self, statement: str) -> Any:
        """Execute a statement (to be implemented by executor)."""
        if self.dry_run:
            logger.info(f"[DRY RUN] Would execute: {statement}")
            return None
        
        logger.info(f"Executing: {statement}")
        return None


@dataclass
class MigrationConfig:
    """Migration configuration."""
    table_name: str = "_migrations"
    lock_timeout_seconds: int = 60
    allow_out_of_order: bool = False
    validate_checksums: bool = True
    baseline_version: Optional[str] = None


class MigrationStore(ABC):
    """Abstract migration store."""
    
    @abstractmethod
    async def get_applied_migrations(self) -> List[MigrationRecord]:
        """Get all applied migrations."""
        pass
    
    @abstractmethod
    async def save_migration(
        self,
        record: MigrationRecord,
    ) -> None:
        """Save migration record."""
        pass
    
    @abstractmethod
    async def get_current_version(self) -> Optional[str]:
        """Get current version."""
        pass
    
    @abstractmethod
    async def delete_migration(
        self,
        migration_id: str,
    ) -> bool:
        """Delete migration record."""
        pass


class InMemoryMigrationStore(MigrationStore):
    """In-memory migration store."""
    
    def __init__(self):
        self._migrations: Dict[str, MigrationRecord] = {}
        self._lock = asyncio.Lock()
    
    async def get_applied_migrations(self) -> List[MigrationRecord]:
        return sorted(
            self._migrations.values(),
            key=lambda m: m.applied_at,
        )
    
    async def save_migration(
        self,
        record: MigrationRecord,
    ) -> None:
        async with self._lock:
            self._migrations[record.migration_id] = record
    
    async def get_current_version(self) -> Optional[str]:
        migrations = await self.get_applied_migrations()
        
        completed = [
            m for m in migrations
            if m.status == MigrationStatus.COMPLETED
        ]
        
        if completed:
            return completed[-1].version
        
        return None
    
    async def delete_migration(
        self,
        migration_id: str,
    ) -> bool:
        async with self._lock:
            if migration_id in self._migrations:
                del self._migrations[migration_id]
                return True
            return False


class MigrationRegistry:
    """Registry for migration definitions."""
    
    def __init__(self):
        self._migrations: Dict[str, MigrationDefinition] = {}
    
    def register(
        self,
        definition: MigrationDefinition,
    ) -> None:
        """Register a migration."""
        self._migrations[definition.migration_id] = definition
    
    def get(
        self,
        migration_id: str,
    ) -> Optional[MigrationDefinition]:
        """Get migration by ID."""
        return self._migrations.get(migration_id)
    
    def get_by_version(
        self,
        version: str,
    ) -> List[MigrationDefinition]:
        """Get migrations for a version."""
        return [
            m for m in self._migrations.values()
            if m.version == version
        ]
    
    def get_all(self) -> List[MigrationDefinition]:
        """Get all migrations sorted by version."""
        return sorted(
            self._migrations.values(),
            key=lambda m: Version.parse(m.version),
        )
    
    def get_pending(
        self,
        applied: Set[str],
    ) -> List[MigrationDefinition]:
        """Get pending migrations."""
        return [
            m for m in self.get_all()
            if m.migration_id not in applied
        ]


class Migrator:
    """
    Database/schema migrator.
    """
    
    def __init__(
        self,
        store: MigrationStore,
        registry: MigrationRegistry,
        config: Optional[MigrationConfig] = None,
    ):
        self._store = store
        self._registry = registry
        self._config = config or MigrationConfig()
        self._lock = asyncio.Lock()
    
    @property
    def registry(self) -> MigrationRegistry:
        return self._registry
    
    async def get_current_version(self) -> Optional[str]:
        """Get current version."""
        return await self._store.get_current_version()
    
    async def get_pending_migrations(self) -> List[MigrationDefinition]:
        """Get pending migrations."""
        applied = await self._store.get_applied_migrations()
        applied_ids = {m.migration_id for m in applied}
        
        return self._registry.get_pending(applied_ids)
    
    async def get_applied_migrations(self) -> List[MigrationRecord]:
        """Get applied migrations."""
        return await self._store.get_applied_migrations()
    
    async def migrate(
        self,
        dry_run: bool = False,
    ) -> List[MigrationRecord]:
        """Run all pending migrations."""
        async with self._lock:
            pending = await self.get_pending_migrations()
            results = []
            
            for definition in pending:
                try:
                    record = await self._run_migration(
                        definition,
                        MigrationDirection.UP,
                        dry_run,
                    )
                    results.append(record)
                    
                    if record.status == MigrationStatus.FAILED:
                        break
                
                except Exception as e:
                    logger.error(f"Migration failed: {e}")
                    break
            
            return results
    
    async def migrate_to(
        self,
        target_version: str,
        dry_run: bool = False,
    ) -> List[MigrationRecord]:
        """Migrate to specific version."""
        if target_version == "latest":
            return await self.migrate(dry_run)
        
        async with self._lock:
            current = await self.get_current_version()
            
            if current:
                current_v = Version.parse(current)
            else:
                current_v = Version(0, 0, 0)
            
            target_v = Version.parse(target_version)
            
            if target_v > current_v:
                # Migrate up
                pending = await self.get_pending_migrations()
                migrations = [
                    m for m in pending
                    if Version.parse(m.version) <= target_v
                ]
            else:
                # Migrate down
                applied = await self.get_applied_migrations()
                migrations = [
                    self._registry.get(m.migration_id)
                    for m in reversed(applied)
                    if Version.parse(m.version) > target_v
                ]
                return await self._rollback_migrations(
                    [m for m in migrations if m],
                    dry_run,
                )
            
            results = []
            
            for definition in migrations:
                record = await self._run_migration(
                    definition,
                    MigrationDirection.UP,
                    dry_run,
                )
                results.append(record)
                
                if record.status == MigrationStatus.FAILED:
                    break
            
            return results
    
    async def rollback(
        self,
        steps: int = 1,
        dry_run: bool = False,
    ) -> List[MigrationRecord]:
        """Rollback last N migrations."""
        async with self._lock:
            applied = await self.get_applied_migrations()
            
            if not applied:
                return []
            
            to_rollback = list(reversed(applied))[:steps]
            
            return await self._rollback_migrations(
                [
                    self._registry.get(m.migration_id)
                    for m in to_rollback
                    if self._registry.get(m.migration_id)
                ],
                dry_run,
            )
    
    async def _rollback_migrations(
        self,
        migrations: List[MigrationDefinition],
        dry_run: bool,
    ) -> List[MigrationRecord]:
        """Rollback a list of migrations."""
        results = []
        
        for definition in migrations:
            if not definition.down:
                raise RollbackError(
                    f"No down migration for {definition.migration_id}"
                )
            
            record = await self._run_migration(
                definition,
                MigrationDirection.DOWN,
                dry_run,
            )
            results.append(record)
            
            if record.status == MigrationStatus.FAILED:
                break
            
            # Delete migration record
            if not dry_run:
                await self._store.delete_migration(definition.migration_id)
        
        return results
    
    async def _run_migration(
        self,
        definition: MigrationDefinition,
        direction: MigrationDirection,
        dry_run: bool,
    ) -> MigrationRecord:
        """Run a single migration."""
        import time
        
        record = MigrationRecord(
            migration_id=definition.migration_id,
            version=definition.version,
            name=definition.name,
            status=MigrationStatus.RUNNING,
            checksum=definition.checksum,
        )
        
        if not dry_run:
            await self._store.save_migration(record)
        
        ctx = MigrationContext(
            migration_id=definition.migration_id,
            version=definition.version,
            direction=direction,
            dry_run=dry_run,
        )
        
        start = time.time()
        
        try:
            if direction == MigrationDirection.UP:
                if asyncio.iscoroutinefunction(definition.up):
                    await definition.up(ctx)
                else:
                    definition.up(ctx)
            else:
                if definition.down:
                    if asyncio.iscoroutinefunction(definition.down):
                        await definition.down(ctx)
                    else:
                        definition.down(ctx)
            
            record.status = MigrationStatus.COMPLETED
            record.completed_at = datetime.now()
            record.execution_time_ms = int((time.time() - start) * 1000)
            
            if not dry_run:
                await self._store.save_migration(record)
            
            logger.info(
                f"Migration {definition.migration_id} "
                f"({direction.value}) completed in {record.execution_time_ms}ms"
            )
        
        except Exception as e:
            record.status = MigrationStatus.FAILED
            record.error = str(e)
            record.execution_time_ms = int((time.time() - start) * 1000)
            
            if not dry_run:
                await self._store.save_migration(record)
            
            logger.error(
                f"Migration {definition.migration_id} failed: {e}"
            )
        
        return record
    
    async def validate(self) -> List[str]:
        """Validate migrations."""
        issues = []
        
        applied = await self.get_applied_migrations()
        
        for record in applied:
            definition = self._registry.get(record.migration_id)
            
            if not definition:
                issues.append(
                    f"Applied migration {record.migration_id} not found in registry"
                )
                continue
            
            if self._config.validate_checksums and record.checksum:
                if record.checksum != definition.checksum:
                    issues.append(
                        f"Checksum mismatch for {record.migration_id}"
                    )
        
        return issues


class MigrationBuilder:
    """
    Builder for creating migrations.
    """
    
    def __init__(self, version: str):
        self._version = version
        self._migrations: List[MigrationDefinition] = []
        self._up_statements: List[str] = []
        self._down_statements: List[str] = []
    
    def add_up(self, statement: str) -> "MigrationBuilder":
        """Add up statement."""
        self._up_statements.append(statement)
        return self
    
    def add_down(self, statement: str) -> "MigrationBuilder":
        """Add down statement."""
        self._down_statements.append(statement)
        return self
    
    def create_table(
        self,
        table_name: str,
        columns: Dict[str, str],
    ) -> "MigrationBuilder":
        """Add create table migration."""
        cols = ", ".join(f"{name} {dtype}" for name, dtype in columns.items())
        self._up_statements.append(
            f"CREATE TABLE {table_name} ({cols})"
        )
        self._down_statements.append(f"DROP TABLE {table_name}")
        return self
    
    def add_column(
        self,
        table_name: str,
        column_name: str,
        column_type: str,
    ) -> "MigrationBuilder":
        """Add column migration."""
        self._up_statements.append(
            f"ALTER TABLE {table_name} ADD COLUMN {column_name} {column_type}"
        )
        self._down_statements.append(
            f"ALTER TABLE {table_name} DROP COLUMN {column_name}"
        )
        return self
    
    def drop_column(
        self,
        table_name: str,
        column_name: str,
    ) -> "MigrationBuilder":
        """Add drop column migration."""
        self._up_statements.append(
            f"ALTER TABLE {table_name} DROP COLUMN {column_name}"
        )
        return self
    
    def add_index(
        self,
        table_name: str,
        index_name: str,
        columns: List[str],
        unique: bool = False,
    ) -> "MigrationBuilder":
        """Add index migration."""
        unique_str = "UNIQUE " if unique else ""
        cols = ", ".join(columns)
        self._up_statements.append(
            f"CREATE {unique_str}INDEX {index_name} ON {table_name} ({cols})"
        )
        self._down_statements.append(f"DROP INDEX {index_name}")
        return self
    
    def build(self, name: str) -> MigrationDefinition:
        """Build migration definition."""
        migration_id = f"{self._version}_{name}"
        
        up_stmts = self._up_statements.copy()
        down_stmts = list(reversed(self._down_statements))
        
        async def up(ctx: MigrationContext):
            for stmt in up_stmts:
                await ctx.execute(stmt)
        
        async def down(ctx: MigrationContext):
            for stmt in down_stmts:
                await ctx.execute(stmt)
        
        checksum = hashlib.md5(
            "".join(up_stmts).encode()
        ).hexdigest()
        
        return MigrationDefinition(
            migration_id=migration_id,
            version=self._version,
            name=name,
            up=up,
            down=down if down_stmts else None,
            checksum=checksum,
        )


# Global registry
_global_registry = MigrationRegistry()


# Decorators
def migration(
    version: str,
    name: Optional[str] = None,
    dependencies: Optional[List[str]] = None,
) -> Callable:
    """
    Decorator to define a migration.
    
    Example:
        @migration(version="1.0.0")
        async def create_users_table(ctx):
            await ctx.execute("CREATE TABLE users ...")
    """
    def decorator(func: Callable) -> Callable:
        migration_name = name or func.__name__
        migration_id = f"{version}_{migration_name}"
        
        # Compute checksum from function source
        import inspect
        source = inspect.getsource(func)
        checksum = hashlib.md5(source.encode()).hexdigest()
        
        definition = MigrationDefinition(
            migration_id=migration_id,
            version=version,
            name=migration_name,
            up=func,
            dependencies=dependencies or [],
            checksum=checksum,
        )
        
        _global_registry.register(definition)
        
        # Store definition on function
        func._migration = definition
        
        return func
    
    return decorator


def rollback(up_func: Callable) -> Callable:
    """
    Decorator to define rollback for a migration.
    
    Example:
        @migration(version="1.0.0")
        async def create_users_table(ctx):
            await ctx.execute("CREATE TABLE users ...")
        
        @rollback(create_users_table)
        async def rollback_users_table(ctx):
            await ctx.execute("DROP TABLE users")
    """
    def decorator(func: Callable) -> Callable:
        if hasattr(up_func, "_migration"):
            up_func._migration.down = func
        
        return func
    
    return decorator


def versioned(
    version: str,
) -> Callable:
    """
    Decorator to mark a function as version-specific.
    
    Example:
        @versioned("2.0.0")
        def new_api_handler():
            ...
    """
    def decorator(func: Callable) -> Callable:
        func._version = version
        
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            return func(*args, **kwargs)
        
        return wrapper
    
    return decorator


# Factory functions
def create_migrator(
    store: Optional[MigrationStore] = None,
    registry: Optional[MigrationRegistry] = None,
) -> Migrator:
    """Create a migrator."""
    s = store or InMemoryMigrationStore()
    r = registry or _global_registry
    return Migrator(s, r)


def create_migration_builder(
    version: str,
) -> MigrationBuilder:
    """Create a migration builder."""
    return MigrationBuilder(version)


def create_migration(
    version: str,
    name: str,
    up: Callable,
    down: Optional[Callable] = None,
) -> MigrationDefinition:
    """Create a migration definition."""
    migration_id = f"{version}_{name}"
    
    return MigrationDefinition(
        migration_id=migration_id,
        version=version,
        name=name,
        up=up,
        down=down,
    )


def get_global_registry() -> MigrationRegistry:
    """Get global migration registry."""
    return _global_registry


__all__ = [
    # Exceptions
    "MigrationError",
    "RollbackError",
    "VersionConflictError",
    # Enums
    "MigrationStatus",
    "MigrationDirection",
    # Data classes
    "Version",
    "MigrationRecord",
    "MigrationDefinition",
    "MigrationContext",
    "MigrationConfig",
    # Stores
    "MigrationStore",
    "InMemoryMigrationStore",
    # Core classes
    "MigrationRegistry",
    "Migrator",
    "MigrationBuilder",
    # Decorators
    "migration",
    "rollback",
    "versioned",
    # Factory functions
    "create_migrator",
    "create_migration_builder",
    "create_migration",
    "get_global_registry",
]
